#include "buffer.h"
#include "server.h"
#include "log.h"
#include "connections.h"
#include "fdevent.h"

#include "request.h"
#include "response.h"
#include "network.h"
#include "http_chunk.h"
#include "stat_cache.h"
#include "joblist.h"

#include "plugin.h"

#include "inet_ntop_cache.h"

/*Added by rzhang*/
#ifdef BUILD_ON_ARM
    #include "nvram.h"
#endif

#define UPLOAD_TOTAL ":up_total"
#define UPLOAD_RECV  ":up_receive"
#define UPLOAD_PROGRESS ":up_progress"
int s_setting_nvram = 0;
/*Added end*/

#include <sys/stat.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <assert.h>
#include <signal.h>
#include <dirent.h>

#ifdef USE_OPENSSL
    #include <openssl/ssl.h>
    #include <openssl/err.h>
#endif

#ifdef HAVE_SYS_FILIO_H
    #include <sys/filio.h>
#endif

#include "sys-socket.h"

typedef struct
{
    PLUGIN_DATA;
} plugin_data;

#if 0
/* Add upload process, from GXV3175 lighttpd codes, by vanevan */
#ifdef BUILD_ON_ARM
#define SOUNDS_JSON_FILE                   "/cfg/var/lib/asterisk/static-http/config/userdefined/language.json"
#define UPLOAD_SOUNDS_PATH                 "/data/var/lib/asterisk/sounds"
#else
#define SOUNDS_JSON_FILE                   "/usr/local/var/lib/asterisk/static-http/config/userdefined/language.json"
#define UPLOAD_SOUNDS_PATH                 "/usr/local/var/lib/asterisk/sounds"
#endif
#define UPLOAD_SOUNDS_TMP_PATH             "/tmp/lang"
#define UPLOAD_ASTERISK_TLS_PATH           "/cfg/etc/asterisk/keys"
#define UPLOAD_ASTERISK_TLS_CA_FILE      	"ca.crt"
#define UPLOAD_ASTERISK_TLS_CERT_FILE      "asterisk.pem"

#define UPLOAD_TYPE			"type"
#define UPLOAD_DATA			"data"
#define UPLOAD_TYPE_IVR			"ivr"
#define UPLOAD_TYPE_MOH			"moh"
#define UPLOAD_TYPE_BACKUP		        "backup"
#define UPLOAD_TYPE_FIRMWARE		"firmware"
#define UPLOAD_TYPE_FLASHUPDATE		"flashupdate"
#define UPLOAD_TYPE_TONEZONE		"tonezone"
#define UPLOAD_TYPE_OVERLAY		"overlay"
#define UPLOAD_TYPE_8021X_CA_CERT      	"8021x_ca_cert"
#define UPLOAD_TYPE_8021X_CLIENT_CERT      "8021x_client_cert"
#define UPLOAD_TYPE_VOICE_PACKEAGE      	"voice_package"
#define UPLOAD_TYPE_ASTERISK_TLS_CA      	"sip_tls_ca"
#define UPLOAD_TYPE_ASTERISK_TLS_CERT      "sip_tls_cert"

#define UPLOAD_RES_OK			"Result=0\r\nMessage=OK\r\n"
#define UPLOAD_RES_NOT_READY		"Result=1\r\nMessage=Not ready yet\r\n"
#define UPLOAD_RES_SERVER_ERROR            "Result=1\r\nMessage=Server error\r\n"
#define UPLOAD_RES_PROVISION_RUNNING       "Result=1\r\nMessage=Upgrade has been started, please wait and try again\r\n"
#define UPLOAD_RES_MISSING_TYPE            "Result=1\r\nMessage=Missing type\r\n"
#define UPLOAD_RES_UNSUPPORTED_TYPE        "Result=1\r\nMessage=Unsupported type\r\n"

#define TEMP_PATH 				        "/tmp"
#define PROVISION_LOCK 			        "/tmp/.provision.lock"
#define PROVISION_PROG			        "/sbin/provision_loadfirmware"

/* Add Info process, from getting device info without login, by rzhang */
#define INFO_TYPE_MODEL_INFO     "model_info"

static char *process_voice_package(server *srv, const char *filename);

/*Added by rzhang, for remove and add language info to language.json file*/
static int check_remove_data_from_json(server *srv, const char *id)
{
    int ret = -1;
    if ( id == NULL || strcmp(id, "en") == 0 )/*Don't allow to remove the en package*/
    {
        log_error_write(srv, __FILE__, __LINE__, "s", "Language information is invalid!");
        return -1;
    }
    char line[128] = {0};
    char file_buffer[1024] = {0};
    char complete_id[16] = {0};

    snprintf(complete_id, sizeof(complete_id), "\"%s\",", id);

    FILE *ptr = fopen(SOUNDS_JSON_FILE, "a+");
    if ( ptr != NULL )
    {
        log_error_write(srv, __FILE__, __LINE__, "ss", "Open success!", SOUNDS_JSON_FILE);

        fseek(ptr, 0, SEEK_SET);//start from the start of file
        while ( fgets(line, sizeof(line)-1, ptr) != NULL )
        {
            log_error_write(srv, __FILE__, __LINE__, "ss", "LINE:", line);

            if ( line != NULL && strstr(line, complete_id) != NULL )
            {
                log_error_write(srv, __FILE__, __LINE__, "s", "Remove this line");
                ret = 0; 
            }
            else
            {
                strncat(file_buffer, line, ( sizeof(file_buffer) - 1 ) - strlen(file_buffer));
            }
            memset(line, 0, sizeof(line));
        }

        if ( ret == 0 )
        {
            log_error_write(srv, __FILE__, __LINE__, "s", "Save changes to file!!!");
            fclose(ptr);
            ptr = fopen(SOUNDS_JSON_FILE, "w+");
            fwrite(file_buffer, strlen(file_buffer), 1, ptr);
            //make sure all data are saved.
            fflush(ptr);
        }
        fclose(ptr);
        ptr = NULL;

    }

    return ret;
}

static int add_data_to_json(server *srv, const char *id, const char *lang)
{
    log_error_write(srv, __FILE__, __LINE__, "s", "Enter add_data_to_json!");

    int ret = -1;
    if ( id == NULL || lang == NULL )
    {
        log_error_write(srv, __FILE__, __LINE__, "s", "Language information is invalid!");
        return -1;
    }

    char file_buffer[1024] = {0};
    char line[128] = {0};
    char complete_id[16] = {0};

    snprintf(complete_id, sizeof(complete_id), "\"%s\",", id);

    FILE *ptr = fopen(SOUNDS_JSON_FILE, "a+");
    if ( ptr != NULL )
    {
        log_error_write(srv, __FILE__, __LINE__, "ss", "Open json success", SOUNDS_JSON_FILE);

        fseek(ptr, 0, SEEK_SET);//start from the start of file
        while ( fgets(line, sizeof(line)-1, ptr) != NULL )
        {
            log_error_write(srv, __FILE__, __LINE__, "ss", "LINE:", line);

            if ( line != NULL && strstr(line, complete_id) != NULL )
            {
                log_error_write(srv, __FILE__, __LINE__, "s", "Already have this id!!!!");
                break;
            }

            if ( line != NULL && strstr(line, "]") != NULL )
            {
                /*The last line, going to add the new language*/
                char buf[256] = {0};
                snprintf(buf, sizeof(buf), ",{\"id\": \"%s\", \"name\": \"%s\"}\n", id, lang);
                log_error_write(srv, __FILE__, __LINE__, "ss", "buf:", buf);

                strncat(file_buffer, buf, sizeof(file_buffer) - 1 - strlen(file_buffer));
                ret = 0;
            }
            strncat(file_buffer, line, sizeof(file_buffer) - 1 - strlen(file_buffer));

            memset(line, 0, sizeof(line));
        }

        if ( ret == 0 )
        {
            fclose(ptr);
            ptr = fopen(SOUNDS_JSON_FILE, "w+");
            log_error_write(srv, __FILE__, __LINE__, "s", "Save changes to file!!!");
            fwrite(file_buffer, strlen(file_buffer), 1, ptr);
            //make sure all data are saved.
            fflush(ptr);
        }
        fclose(ptr);
        ptr = NULL;

    }
    else
    {
        log_error_write(srv, __FILE__, __LINE__, "sss", "Open ", SOUNDS_JSON_FILE, "failed !!!!");
    }
    
    log_error_write(srv, __FILE__, __LINE__, "s", "Exit add_data_to_json!");

    return ret;
}
/*Added end*/

static int is_provision_running(server *srv)
{
    FILE *file = NULL;
    char buf[64] = {0};
    int pid = 0;
    int ret = 0;

    file = fopen(PROVISION_LOCK, "r");
    if ( file != NULL )
    {
        if ( fgets(buf, sizeof(buf) - 1, file) != NULL )
        {
            pid = atoi(buf);
            log_error_write(srv, __FILE__, __LINE__, "sd",
                            "Check provision ", pid);
            if ( kill(pid, 0) == 0 )
            {
                log_error_write(srv, __FILE__, __LINE__, "sds",
                                "Provision ", pid, " is running");
                ret = 1;
            }
        }
        fclose(file);
    }

    return ret;
}

static const char *msg_get_header(const struct message *m, const char *var)
{
    char cmp[80];
    unsigned int x;

    snprintf(cmp, sizeof(cmp), "%s=", var);
    for ( x = 0; x < m->hdrcount; x++ )
    {
        if ( !strncasecmp(cmp, m->headers[x], strlen(cmp)) )
        {
            return m->headers[x] + strlen(cmp);
        }
    }

    return NULL;
}

static int generate_file_name(server *srv, connection *con, buffer *b, const struct message *m, char *filename, int len)
{
    int ret = 0;
    const char *type = NULL;
    const char *data = NULL;
    chunk *c = NULL;
    char buf[256] = {0};
    char name[256] = {0};
    int found = 0;

    if ( srv == NULL || con == NULL || b == NULL || m == NULL )
    {
        return -1;
    }

    type = msg_get_header(m, "type");  /* Upload type */
    data = msg_get_header(m, "data");  /* The destination path */
    c = con->request_content_queue->first;
    log_error_write(srv, __FILE__, __LINE__, "sds", "Upload type:", 
                    c != NULL ? (int)c->type : -1, 
                    c != NULL && c->file.name->ptr != NULL ? c->file.name->ptr : "->empty");
    if ( c!= NULL && c->type == FILE_CHUNK && c->file.name->ptr != NULL )  /* Filename */
    {
        FILE *file = NULL;
        const char *ptr = NULL, *end = NULL, *temp = NULL;

        log_error_write(srv, __FILE__, __LINE__, "ssd", "Upload filename:", c->file.name->ptr, c->file.name->size);
        /* Try to get the form data header information,
         * like this: "Content-Disposition: form-data; name="filename"; filename="rcvr.img""
         */
        file = fopen(c->file.name->ptr, "rb");
        if ( file != NULL )
        {
            while ( !feof(file) )
            {
                memset(buf, 0, sizeof(buf));
                fgets(buf, sizeof(buf) - 1, file);
                if ( (ptr = strstr(buf, "Content-Disposition:")) != NULL )
                {
#if 0
                    /* Get "name" property, it contains the real filename */
                    ptr = strstr(ptr, "name=");
                    if ( ptr != NULL )
                    {
                        ptr += strlen("name=");
                        temp = strchr(ptr, '"');
                        end = strchr(ptr, ';');
                        if ( (temp != NULL && end != NULL && (temp <= end)) ||
                             (temp != NULL && end == NULL) )
                        {
                            ++temp;
                            end = strchr(temp, '"');
                            if ( end != NULL )
                            {
                                strncpy(name, temp, end - temp);
                            }
                            else
                            {
                                strcpy(name, temp);
                            }
                        }
                        else if ( (temp != NULL && end != NULL && (temp > end)) ||
                                  (temp == NULL && end != NULL) )
                        {
                            strncpy(name, ptr, end - ptr);
                        }
                        else
                        {
                            strcpy(name, ptr);
                        }
                    }
                    else
                    {
                        strcpy(name, "filename");
                    }
#endif

                    /*Get the real filename*/
                    ptr = strstr(buf, "filename=");
                    if ( ptr != NULL )
                    {
                        ptr += strlen("filename=");
                        temp = strstr(ptr, "\"");
                        end = strstr(ptr, ";");
                        if ( (temp != NULL && end != NULL && (temp <= end)) ||
                             (temp != NULL && end == NULL) )
                        {
                            ++temp;
                            end = strstr(temp, "\"");
                            if ( end != NULL )
                            {
                                strncpy(name, temp, end - temp);
                            }
                            else
                            {
                                strcpy(name, temp);
                            }
                        }
                        else if ( (temp != NULL && end != NULL && (temp > end)) ||
                                  (temp == NULL && end != NULL) )
                        {
                            strncpy(name, ptr, end - ptr);
                        }
                        else
                        {
                            strcpy(name, ptr);
                        }
#if 0
                        /* Delete the append '=' */
                        if ( *(name + strlen(name) - 1) == '=' )
                        {
                            *(name + strlen(name) - 1) = '\0';
                        }
#endif
                        found = 1;  /* Set found flag */
                    }
                    break;
                }

                if ( strstr(buf, "\r\n\r\n") != NULL )
                {
                    break;
                }
            }
            fclose(file);
        }
    }
    else if ( c != NULL && c->type == MEM_CHUNK )
    {
        const char *ptr = NULL, *end = NULL, *temp = NULL;
        log_error_write(srv, __FILE__, __LINE__, "sdd", "Mem chunk:", c->mem->used, c->mem->size);
        if ( (ptr = strstr(c->mem->ptr, "Content-Disposition:")) != NULL )
        {
            /*Get the real filename*/
            ptr = strstr(ptr, "filename=");
            if ( ptr != NULL )
            {
                ptr += strlen("filename=");
                temp = strstr(ptr, "\"");
                end = strstr(ptr, ";");
                if ( (temp != NULL && end != NULL && (temp <= end)) ||
                     (temp != NULL && end == NULL) )
                {
                    ++temp;
                    end = strstr(temp, "\"");
                    if ( end != NULL )
                    {
                        strncpy(name, temp, end - temp);
                    }
                    else
                    {
                        strcpy(name, temp);
                    }
                }
                else if ( (temp != NULL && end != NULL && (temp > end)) ||
                          (temp == NULL && end != NULL) )
                {
                    strncpy(name, ptr, end - ptr);
                }
                else
                {
                    strcpy(name, ptr);
                }
                found = 1;  /* Set found flag */
            }
        }
    }

    /* Not found, use the temp name */
    if ( found == 0 )
    {
        snprintf(name, sizeof(name) - 1, "tmp%ld-%ld", random(), time(NULL));
    }

    if ( type == NULL )
    {
        buffer_append_string(b, UPLOAD_RES_MISSING_TYPE);
        ret = -1;
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_FIRMWARE) == 0 )
    {
        snprintf(filename, len - 1, "%s/%s", 
                 data != NULL ? data : "/update",
                 name);
        log_error_write(srv, __FILE__, __LINE__, "ss", "Firmware filename:", filename);
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_IVR) == 0 )
    {
        if ( data != NULL )
        {
            snprintf(filename, len - 1, "%s/%s", 
                     data, name);
        }
        else
        {
            snprintf(filename, len - 1, "%s/%s/%s",
                     getenv("AST_BASE") != NULL ? getenv("AST_BASE") : "",
                     "/var/lib/asterisk/sounds/record", name);
        }
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_TONEZONE) == 0 )
    {
        buffer_append_string(b, UPLOAD_RES_NOT_READY);
        ret = -1;
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_OVERLAY) == 0 )
    {
        buffer_append_string(b, UPLOAD_RES_NOT_READY);
        ret = -1;
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_BACKUP) == 0 )
    {
        if ( data != NULL )
        {
            snprintf(filename, len - 1, "%s/%s", 
                     data, name);
        }
        else
        {
            snprintf(filename, len - 1, "%s/%s/%s",
                     getenv("AST_BASE") != NULL ? getenv("AST_BASE") : "",
                     "/var/lib/asterisk/gui_backups", name);
        }
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_MOH) == 0 )
    {
        if ( data != NULL )
        {
            snprintf(filename, len - 1, "%s/%s", 
                     data, name);
        }
        else
        {
            snprintf(filename, len - 1, "%s/%s/%s",
                     getenv("AST_BASE") != NULL ? getenv("AST_BASE") : "",
                     "/var/lib/asterisk/moh", name);
        }
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_FLASHUPDATE) == 0 )
    {
        buffer_append_string(b, UPLOAD_RES_NOT_READY);
        ret = -1;
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_8021X_CA_CERT) == 0 )
    {
        if ( data != NULL )
        {
            snprintf(filename, len - 1, "%s/%s", 
                     data, name);
        }
        else
        {
            snprintf(filename, len - 1, "%s",
                     "/cfg/etc/8021x_ca_cert");
        }
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_8021X_CLIENT_CERT) == 0 )
    {
        if ( data != NULL )
        {
            snprintf(filename, len - 1, "%s/%s", 
                     data, name);
        }
        else
        {
            snprintf(filename, len - 1, "%s",
                     "/cfg/etc/8021x_client_cert");
        }
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_ASTERISK_TLS_CA) == 0 )
    {
        if ( data != NULL )
        {
            snprintf(filename, len - 1, "%s/%s", 
                     data, name);
        }
        else
        {
            snprintf(filename, len - 1, "%s/%s",
                     UPLOAD_ASTERISK_TLS_PATH, UPLOAD_ASTERISK_TLS_CA_FILE);
        }
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_ASTERISK_TLS_CERT) == 0 )
    {
        if ( data != NULL )
        {
            snprintf(filename, len - 1, "%s/%s", 
                     data, name);
        }
        else
        {
            snprintf(filename, len - 1, "%s/%s",
                     UPLOAD_ASTERISK_TLS_PATH, UPLOAD_ASTERISK_TLS_CERT_FILE);
        }
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_VOICE_PACKEAGE) == 0 )
    {
        int action_type = 0;/*0 is upload, 1 is delete*/
        if ( con != NULL && con->physical.rel_path != NULL 
             && con->physical.rel_path->ptr != NULL 
             && strcasecmp(con->physical.rel_path->ptr, "/upload") == 0 )
        {
            log_error_write(srv, __FILE__, __LINE__, "ss", "Clean the dir %s before upload!", UPLOAD_SOUNDS_TMP_PATH);

            memset(buf, 0, sizeof(buf));
            snprintf(buf, sizeof(buf), "%s%s%s%s", "rm -rf ", UPLOAD_SOUNDS_TMP_PATH, " && mkdir ", UPLOAD_SOUNDS_TMP_PATH);
            system(buf);
        }
        else
        {
            log_error_write(srv, __FILE__, __LINE__, "s", "Delete language packet!");

            action_type = 1;
        }
        
        if ( data != NULL )
        {
            snprintf(filename, len - 1, "%s/%s", 
                     data, name);
        }
        else
        {
            if ( action_type == 0 )
            {
                snprintf(filename, len - 1, "%s/%s",
                         UPLOAD_SOUNDS_TMP_PATH, name);
            }
            else
            {
                snprintf(filename, len - 1, "%s/%s",
                         UPLOAD_SOUNDS_PATH, name);
            }
        }
    }
    else
    {
        buffer_append_string(b, UPLOAD_RES_UNSUPPORTED_TYPE);
        ret = -1;
    }

    return ret;
}

static int prepare_upload(server *srv, connection *con, buffer *b, const struct message *m)
{
    int ret = 0;
    const char *type = NULL;
    const char *data = NULL;

    if ( srv == NULL || con == NULL || b == NULL || m == NULL )
    {
        buffer_append_string(b, UPLOAD_RES_SERVER_ERROR);
        return -1;
    }

    type = msg_get_header(m, "type");  /* Upload type */
    data = msg_get_header(m, "data");  /* The destination path */
    /* If it is firmware upgrade, check if provision is running first */
    if ( (type != NULL) && (strcasecmp(type, UPLOAD_TYPE_FIRMWARE) == 0) && (is_provision_running(srv) != 0) )
    {
        buffer_append_string(b, UPLOAD_RES_PROVISION_RUNNING);
        return -2;
    }

    if ( type == NULL )
    {
        buffer_append_string(b, UPLOAD_RES_MISSING_TYPE);
        ret = -1;
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_FIRMWARE) == 0 )
    {
        /* Clear update partition to make sure there is enough space */
        system("rm /update/* -rf");
        buffer_append_string(b, UPLOAD_RES_OK);
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_IVR) == 0 )
    {
        buffer_append_string(b, UPLOAD_RES_OK);
        ret = 0;
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_TONEZONE) == 0 )
    {
        buffer_append_string(b, UPLOAD_RES_NOT_READY);
        ret = -1;
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_OVERLAY) == 0 )
    {
        buffer_append_string(b, UPLOAD_RES_NOT_READY);
        ret = -1;
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_BACKUP) == 0 )
    {
        buffer_append_string(b, UPLOAD_RES_OK);
        ret = 0;
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_MOH) == 0 )
    {
        buffer_append_string(b, UPLOAD_RES_OK);
        ret = 0;
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_FLASHUPDATE) == 0 )
    {
        buffer_append_string(b, UPLOAD_RES_NOT_READY);
        ret = -1;
    }
    else if ( strcasecmp(type, UPLOAD_TYPE_VOICE_PACKEAGE) == 0 )
    {
        buffer_append_string(b, UPLOAD_RES_OK);
        ret = -1;
    }
    else
    {
        buffer_append_string(b, UPLOAD_RES_UNSUPPORTED_TYPE);
        ret = -1;
    }

    return ret;

}

static int process_upload(server *srv, connection *con, buffer *b, const struct message *m)
{
    FILE *file_fd = NULL;
    size_t total = 0;
    char temp[32] = "";
    char * start = NULL;
    size_t length = 0;
    char *end = NULL;
    int strip_header = 0;
    char *chunk_start = NULL;
    char *chunk_end = NULL;
    int boundary_length = 29;
    char *boundary_ptr = NULL;
    const char *resType = NULL;
    char *tmp = NULL;
    const char * jsonCallback = NULL;
    int ret = 0;
    char filename[256] = {0};


    if ( con->request.content_length )
    {
        total = con->request.content_length;
        chunkqueue *cq = con->request_content_queue;
        chunk *c;

        assert(chunkqueue_length(cq) == (off_t)con->request.content_length);
        boundary_ptr = strstr(con->request.http_content_type, "boundary=");
        if ( boundary_ptr != NULL )
        {
            boundary_length = strlen(boundary_ptr);
        }

        if ( generate_file_name(srv, con, b, m, filename, sizeof(filename)) != 0 )
        {
            log_error_write(srv, __FILE__, __LINE__, "s", "Generate file failed!\n");
            return -1;
        }
        file_fd = fopen(filename, "wb+");
        if ( file_fd == NULL )
        {
            log_error_write(srv, __FILE__, __LINE__, "sss", "Open ", filename, " failed!\n");
            return -1;
        }
        log_error_write(srv, __FILE__, __LINE__, "sss", "Upload file:", filename, "!\n");

        /* there is content to send */
        for ( c = cq->first; c; c = cq->first )
        {
            int r = 0;
            log_error_write(srv, __FILE__, __LINE__, "sO", "COPY CHUNK LOOP!!!!!!!!!", c->offset);

            strip_header++;
            /* copy all chunks */
            switch ( c->type )
            {
                log_error_write(srv, __FILE__, __LINE__, "sd", "switch COPY CHUNK!!!", c->type);

                case FILE_CHUNK:
                    if ( c->file.mmap.start == MAP_FAILED )
                    {
                        if ( -1 == c->file.fd &&  /* open the file if not already open */
                             -1 == (c->file.fd = open(c->file.name->ptr, O_RDONLY)) )
                        {
                            log_error_write(srv, __FILE__, __LINE__, "ss", "open failed: ", strerror(errno));
                            snprintf(temp,sizeof(temp),"open O_RDONLY failed\r\n");
                            buffer_append_string(b,temp);
                            return -1;
                        }

                        c->file.mmap.length = c->file.length;

                        if ( MAP_FAILED == (c->file.mmap.start = mmap(0,  c->file.mmap.length, PROT_READ, MAP_SHARED, c->file.fd, 0)) )
                        {
                            log_error_write(srv, __FILE__, __LINE__, "ssbd", "mmap failed: ",
                                            strerror(errno), c->file.name,  c->file.fd);
                            snprintf(temp,sizeof(temp),"open PROT_READ failed\r\n");
                            buffer_append_string(b,temp);
                            close(c->file.fd);
                            c->file.fd = -1;
                            return -1;
                        }

                        close(c->file.fd);
                        c->file.fd = -1;

                        /* chunk_reset() or chunk_free() will cleanup for us */
                    }

                    if ( chunk_start != NULL )
                    {
                        free(chunk_start);
                        chunk_start = NULL;
                    }
                    log_error_write(srv, __FILE__, __LINE__, "dd",
                                    c->type, c->offset);
                    log_error_write(srv, __FILE__, __LINE__, "sdddd",
                                    c->file.name->ptr,
                                    c->file.start,
                                    c->file.length,
                                    c->file.mmap.length,
                                    c->file.mmap.offset);

                    chunk_start = malloc(c->file.mmap.length + 1);
                    memset(chunk_start, 0, c->file.mmap.length + 1);
                    memcpy(chunk_start, c->file.mmap.start, c->file.mmap.length);
                    chunk_end = chunk_start + c->file.mmap.length;

                    char *cont = NULL;
                    start = NULL;
                    start = strstr(chunk_start + c->offset, "--");
                    cont = strstr(chunk_start + c->offset, "Content-Type:");
                    if ( (start != NULL) && (cont != NULL) && strip_header == 1 )
                    {
                        start = strstr(cont, "\r\n\r\n");
                        if ( start != NULL )
                        {
                            start += strlen("\r\n\r\n");

                            c->offset += start - chunk_start;
                        }
                        else
                        {
                            start = chunk_start + c->offset;
                        }
                    }
                    else
                    {
                        start = chunk_start + c->offset;
                    }

                    if ( 1 )
                    {
                        end = strstr(chunk_start + c->file.length - boundary_length + 1, "--");
                        if ( end == NULL )
                        {
                            end = chunk_start + c->file.length;
                        }
                        else
                        {
                            char * dectet = NULL;
                            dectet = end -1;

                            while ( ((*dectet == '\r') && (*(dectet + 2) =='-')) 
                                    || ((*dectet == '\n') && (*(dectet + 1) =='-'))
                                    || (*dectet == '-') )
                            {
                                end--;
                                dectet--;
                            } 
                        }
                    }

                    length = end - start;
                    log_error_write(srv, __FILE__, __LINE__, "sd",
                                    "Data length: ", length);

                    if ( (r = fwrite(start, 1, length, file_fd)) <= 0 )
                    {
                        switch ( errno )
                        {
                            case ENOSPC:
                                con->http_status = 507;
                                if ( chunk_start != NULL )
                                {
                                    free(chunk_start);
                                    chunk_start = NULL;
                                }
                                break;
                            default:
                                //con->http_status = 403;
                                if ( chunk_start != NULL )
                                {
                                    free(chunk_start);
                                    chunk_start = NULL;
                                }
                                break;
                        }
                    }
                    if ( chunk_start != NULL )
                    {
                        free(chunk_start);
                        chunk_start = NULL;
                    }
                    break;
                case MEM_CHUNK:

                    if ( (strip_header == 1) && ((start = strstr(c->mem->ptr + c->offset, "\r\n\r\n")) != NULL) 
                         && (strstr(c->mem->ptr + c->offset, "Content-Type:") != NULL) )
                    {
                        start += strlen("\r\n\r\n");
                    }
                    else
                    {
                        start = c->mem->ptr + c->offset;
                    }
                    //end = strstr(c->mem->ptr + c->mem->used - 55, "\r\n--------------------");
                    if ( 1 )
                    {
                        end = strstr(c->mem->ptr + c->mem->used - boundary_length, "--");
                        if ( end == NULL )
                        {
                            end = c->mem->ptr + c->mem->used - 1;
                        }
                        else
                        {
                            char * dectet = NULL;
                            dectet = end -1;
                            //end += strlen("\r\n");

                            while ( ((*dectet == '\r') && (*(dectet + 2) =='-')) 
                                    || ((*dectet == '\n') && (*(dectet + 1) =='-'))
                                    || (*dectet == '-') )
                            {
                                end--;
                                dectet--;
                            } 
                        }
                    }
                    length = end - start;

                    r = c->mem->used - c->offset - 1;
                    if ( (fwrite(start, 1, length, file_fd)) <= 0 )
                    {
                        switch ( errno )
                        {
                            case ENOSPC:
                                con->http_status = 507;

                                break;
                            default:
                                //con->http_status = 403;
                                break;
                        }
                    }
                    break;
                case UNUSED_CHUNK:
                    break;
            }

            if ( r > 0 )
            {
                c->offset += r;
                cq->bytes_out += r;
            }
            else
            {
                break;
            }

            //log_error_write(srv, __FILE__, __LINE__, "sdsO",
            //                         "buffer total: ", total, "buffer received: ", cq->bytes_out );
            //double progress = (int)cq->bytes_out*1.0;
            //log_error_write(srv, __FILE__, __LINE__, "sO",
            //                         "Upload progress: ", (off_t)(100*progress)/total);
            chunkqueue_remove_finished_chunks(cq);
        }
#ifdef BUILD_ON_ARM
        const char *total_ptr = nvram_get(UPLOAD_TOTAL);
        nvram_set(UPLOAD_RECV, total_ptr);
        //nvram_set(UPLOAD_PROGRESS, total_ptr);
#endif

        log_error_write(srv, __FILE__, __LINE__, "s", "COPY CHUNK LOOP END!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");

        //log_error_write(srv, __FILE__, __LINE__, "s",
        //                            "buffer received finished!!!!");
        //log_error_write(srv, __FILE__, __LINE__, "sd",
        //                            "Total is :", total);

        if ( file_fd != NULL )
        {
            fclose(file_fd);
            file_fd = NULL;
        }
    }

    if ( ret == 0 )
    {
        resType = msg_get_header(m, "format");
        if ( (resType != NULL) && !strcasecmp(resType, "json") )
        {
            jsonCallback = msg_get_header( m, "jsoncallback" );    

            if ( jsonCallback != NULL )
            {
                tmp = malloc((strlen(jsonCallback) + 64) * sizeof(char));

                if ( tmp != NULL )
                {
                    sprintf(tmp, "%s(%s)", jsonCallback, "{\"res\" : \"success\", \"msg\" : \"upload down\"}");
                    buffer_append_string(b, tmp);
                    free(tmp);
                }
            }
            else
            {
                buffer_append_string(b, "{\"res\" : \"success\", \"msg\" : \"upload down\"}");
            }
        }
        else
        {
            const char *type = NULL;
            int code = 0;
            char buf[256] = {0};
            type = msg_get_header(m, "type");
            if ( type != NULL && strcasecmp(type, UPLOAD_TYPE_FIRMWARE) == 0 )
            {
                snprintf(buf, sizeof(buf) - 1, "%s %s", PROVISION_PROG, filename);
                code = system(buf);
                log_error_write(srv, __FILE__, __LINE__, "ssd", "Upgrade command:", buf, code);
                if ( code == 0 )
                {
                    snprintf(buf, sizeof(buf) - 1,
                             "Result=%d\r\n", code);
                }
                else
                {
                    snprintf(buf, sizeof(buf) - 1, 
                             "Result=%d\r\n", WEXITSTATUS(code));
                }
                buffer_append_string(b, buf);
            }
            else if ( type != NULL && strcasecmp(type, UPLOAD_TYPE_VOICE_PACKEAGE) == 0 )
            {
                char *output = process_voice_package(srv, filename);
                if ( output == NULL )
                {
                    snprintf(buf, sizeof(buf) - 1, "Result=0\r\n");
                }
                else
                {
                    log_error_write(srv, __FILE__, __LINE__, "sss", "Process Voice package ", filename, output);

                    snprintf(buf, sizeof(buf) - 1, "Result=-1\r\nError=%s\r\n", output);
                    free(output);
                }
                buffer_append_string(b, buf);
            }
            else
            {
                buffer_append_string(b, "Result=0\r\n");
            }

        }
    }

    return ret;
}
/*** Add end ***/
#endif

static connection *connections_get_new_connection(server *srv)
{
    connections *conns = srv->conns;
    size_t i;

    if ( conns->size == 0 )
    {
        conns->size = 128;
        conns->ptr = NULL;
        conns->ptr = malloc(sizeof(*conns->ptr) * conns->size);
        for ( i = 0; i < conns->size; i++ )
        {
            conns->ptr[i] = connection_init(srv);
        }
    }
    else if ( conns->size == conns->used )
    {
        conns->size += 128;
        conns->ptr = realloc(conns->ptr, sizeof(*conns->ptr) * conns->size);

        for ( i = conns->used; i < conns->size; i++ )
        {
            conns->ptr[i] = connection_init(srv);
        }
    }

    connection_reset(srv, conns->ptr[conns->used]);
#if 0
    fprintf(stderr, "%s.%d: add: ", __FILE__, __LINE__);
    for ( i = 0; i < conns->used + 1; i++ )
    {
        fprintf(stderr, "%d ", conns->ptr[i]->fd);
    }
    fprintf(stderr, "\n");
#endif

    conns->ptr[conns->used]->ndx = conns->used;
    return conns->ptr[conns->used++];
}

static int connection_del(server *srv, connection *con)
{
    size_t i;
    connections *conns = srv->conns;
    connection *temp;

    if ( con == NULL ) return -1;

    if ( -1 == con->ndx ) return -1;

    buffer_reset(con->uri.authority);
    buffer_reset(con->uri.path);
    buffer_reset(con->uri.query);
    buffer_reset(con->request.orig_uri);

    i = con->ndx;

    /* not last element */

    if ( i != conns->used - 1 )
    {
        temp = conns->ptr[i];
        conns->ptr[i] = conns->ptr[conns->used - 1];
        conns->ptr[conns->used - 1] = temp;

        conns->ptr[i]->ndx = i;
        conns->ptr[conns->used - 1]->ndx = -1;
    }

    conns->used--;

    con->ndx = -1;
#if 0
    fprintf(stderr, "%s.%d: del: (%d)", __FILE__, __LINE__, conns->used);
    for ( i = 0; i < conns->used; i++ )
    {
        fprintf(stderr, "%d ", conns->ptr[i]->fd);
    }
    fprintf(stderr, "\n");
#endif
    return 0;
}

int connection_close(server *srv, connection *con)
{
#ifdef USE_OPENSSL
    server_socket *srv_sock = con->srv_socket;
#endif

#ifdef USE_OPENSSL
    if ( srv_sock->is_ssl )
    {
        if ( con->ssl ) SSL_free(con->ssl);
        con->ssl = NULL;
    }
#endif

    fdevent_event_del(srv->ev, &(con->fde_ndx), con->fd);
    fdevent_unregister(srv->ev, con->fd);
#ifdef __WIN32
    if ( closesocket(con->fd) )
    {
        log_error_write(srv, __FILE__, __LINE__, "sds",
                        "(warning) close:", con->fd, strerror(errno));
    }
#else
    if ( close(con->fd) )
    {
        log_error_write(srv, __FILE__, __LINE__, "sds",
                        "(warning) close:", con->fd, strerror(errno));
    }
#endif

    srv->cur_fds--;
#if 0
    log_error_write(srv, __FILE__, __LINE__, "sd",
                    "closed()", con->fd);
#endif

    connection_del(srv, con);
    connection_set_state(srv, con, CON_STATE_CONNECT);

    return 0;
}

#if 0
static void dump_packet(const unsigned char *data, size_t len)
{
    size_t i, j;

    if ( len == 0 ) return;

    for ( i = 0; i < len; i++ )
    {
        if ( i % 16 == 0 ) fprintf(stderr, "  ");

        fprintf(stderr, "%02x ", data[i]);

        if ( (i + 1) % 16 == 0 )
        {
            fprintf(stderr, "  ");
            for ( j = 0; j <= i % 16; j++ )
            {
                unsigned char c;

                if ( i-15+j >= len ) break;

                c = data[i-15+j];

                fprintf(stderr, "%c", c > 32 && c < 128 ? c : '.');
            }

            fprintf(stderr, "\n");
        }
    }

    if ( len % 16 != 0 )
    {
        for ( j = i % 16; j < 16; j++ )
        {
            fprintf(stderr, "   ");
        }

        fprintf(stderr, "  ");
        for ( j = i & ~0xf; j < len; j++ )
        {
            unsigned char c;

            c = data[j];
            fprintf(stderr, "%c", c > 32 && c < 128 ? c : '.');
        }
        fprintf(stderr, "\n");
    }
}
#endif

static int connection_handle_read_ssl(server *srv, connection *con)
{
#ifdef USE_OPENSSL
    int r, ssl_err, len, count = 0, read_offset, toread;
    buffer *b = NULL;

    if ( !con->conf.is_ssl ) return -1;

    ERR_clear_error();
    do
    {
        if ( NULL != con->read_queue->last )
        {
            b = con->read_queue->last->mem;
        }

        if ( NULL == b || b->size - b->used < 1024 )
        {
            b = chunkqueue_get_append_buffer(con->read_queue);
            len = SSL_pending(con->ssl);
            if ( len < 4*1024 ) len = 4*1024; /* always alloc >= 4k buffer */
            buffer_prepare_copy(b, len + 1);

            /* overwrite everything with 0 */
            memset(b->ptr, 0, b->size);
        }

        read_offset = (b->used > 0) ? b->used - 1 : 0;
        toread = b->size - 1 - read_offset;

        len = SSL_read(con->ssl, b->ptr + read_offset, toread);

        if ( len > 0 )
        {
            if ( b->used > 0 ) b->used--;
            b->used += len;
            b->ptr[b->used++] = '\0';

            con->bytes_read += len;

            count += len;
        }
    } while ( len == toread && count < MAX_READ_LIMIT );


    if ( len < 0 )
    {
        int oerrno = errno;
        switch ( (r = SSL_get_error(con->ssl, len)) )
        {
            case SSL_ERROR_WANT_READ:
            case SSL_ERROR_WANT_WRITE:
                con->is_readable = 0;

                /* the manual says we have to call SSL_read with the same arguments next time.
                 * we ignore this restriction; no one has complained about it in 1.5 yet, so it probably works anyway.
                 */

                return 0;
            case SSL_ERROR_SYSCALL:
                /**
                 * man SSL_get_error()
                 *
                 * SSL_ERROR_SYSCALL
                 *   Some I/O error occurred.  The OpenSSL error queue may contain more
                 *   information on the error.  If the error queue is empty (i.e.
                 *   ERR_get_error() returns 0), ret can be used to find out more about
                 *   the error: If ret == 0, an EOF was observed that violates the
                 *   protocol.  If ret == -1, the underlying BIO reported an I/O error
                 *   (for socket I/O on Unix systems, consult errno for details).
                 *
                 */
                while ( (ssl_err = ERR_get_error()) )
                {
                    /* get all errors from the error-queue */
                    log_error_write(srv, __FILE__, __LINE__, "sds", "SSL:",
                                    r, ERR_error_string(ssl_err, NULL));
                }

                switch ( oerrno )
                {
                    default:
                        log_error_write(srv, __FILE__, __LINE__, "sddds", "SSL:",
                                        len, r, oerrno,
                                        strerror(oerrno));
                        break;
                }

                break;
            case SSL_ERROR_ZERO_RETURN:
                /* clean shutdown on the remote side */

                if ( r == 0 )
                {
                    /* FIXME: later */
                }

                /* fall thourgh */
            default:
                while ( (ssl_err = ERR_get_error()) )
                {
                    switch ( ERR_GET_REASON(ssl_err) )
                    {
                        case SSL_R_SSL_HANDSHAKE_FAILURE:
                        case SSL_R_TLSV1_ALERT_UNKNOWN_CA:
                        case SSL_R_SSLV3_ALERT_CERTIFICATE_UNKNOWN:
                        case SSL_R_SSLV3_ALERT_BAD_CERTIFICATE:
                            if ( !con->conf.log_ssl_noise ) continue;
                            break;
                        default:
                            break;
                    }
                    /* get all errors from the error-queue */
                    log_error_write(srv, __FILE__, __LINE__, "sds", "SSL:",
                                    r, ERR_error_string(ssl_err, NULL));
                }
                break;
        }

        connection_set_state(srv, con, CON_STATE_ERROR);

        return -1;
    }
    else if ( len == 0 )
    {
        con->is_readable = 0;
        /* the other end close the connection -> KEEP-ALIVE */

        return -2;
    }
    else
    {
        joblist_append(srv, con);
    }

    return 0;
#else
    UNUSED(srv);
    UNUSED(con);
    return -1;
#endif
}

/* 0: everything ok, -1: error, -2: con closed */
static int connection_handle_read(server *srv, connection *con)
{
    int len;
    buffer *b;
    int toread, read_offset;

    if ( con->conf.is_ssl )
    {
        return connection_handle_read_ssl(srv, con);
    }

    b = (NULL != con->read_queue->last) ? con->read_queue->last->mem : NULL;

    /* default size for chunks is 4kb; only use bigger chunks if FIONREAD tells
     *  us more than 4kb is available
     * if FIONREAD doesn't signal a big chunk we fill the previous buffer
     *  if it has >= 1kb free
     */
#if defined(__WIN32)
    if ( NULL == b || b->size - b->used < 1024 )
    {
        b = chunkqueue_get_append_buffer(con->read_queue);
        buffer_prepare_copy(b, 4 * 1024);
    }

    read_offset = (b->used == 0) ? 0 : b->used - 1;
    len = recv(con->fd, b->ptr + read_offset, b->size - 1 - read_offset, 0);
#else
    if ( ioctl(con->fd, FIONREAD, &toread) || toread == 0 || toread <= 4*1024 )
    {
        if ( NULL == b || b->size - b->used < 1024 )
        {
            b = chunkqueue_get_append_buffer(con->read_queue);
            buffer_prepare_copy(b, 4 * 1024);
        }
    }
    else
    {
        if ( toread > MAX_READ_LIMIT ) toread = MAX_READ_LIMIT;
        b = chunkqueue_get_append_buffer(con->read_queue);
        buffer_prepare_copy(b, toread + 1);
    }

    read_offset = (b->used == 0) ? 0 : b->used - 1;
    len = read(con->fd, b->ptr + read_offset, b->size - 1 - read_offset);
#endif

    if ( len < 0 )
    {
        con->is_readable = 0;

        if ( errno == EAGAIN ) return 0;
        if ( errno == EINTR )
        {
            /* we have been interrupted before we could read */
            con->is_readable = 1;
            return 0;
        }

        if ( errno != ECONNRESET )
        {
            /* expected for keep-alive */
            log_error_write(srv, __FILE__, __LINE__, "ssd", "connection closed - read failed: ", strerror(errno), errno);
        }

        connection_set_state(srv, con, CON_STATE_ERROR);

        return -1;
    }
    else if ( len == 0 )
    {
        con->is_readable = 0;
        /* the other end close the connection -> KEEP-ALIVE */

        /* pipelining */

        return -2;
    }
    else if ( (size_t)len < b->size - 1 )
    {
        /* we got less then expected, wait for the next fd-event */

        con->is_readable = 0;
    }

    if ( b->used > 0 ) b->used--;
    b->used += len;
    b->ptr[b->used++] = '\0';

    con->bytes_read += len;
#if 0
    dump_packet(b->ptr, len);
#endif

    return 0;
}

static int connection_handle_write_prepare(server *srv, connection *con)
{

    if ( con->mode == DIRECT )
    {
        /* static files */
        switch ( con->request.http_method )
        {
            case HTTP_METHOD_GET:
            case HTTP_METHOD_POST:
            case HTTP_METHOD_HEAD:
            case HTTP_METHOD_PUT:
            case HTTP_METHOD_MKCOL:
            case HTTP_METHOD_DELETE:
            case HTTP_METHOD_COPY:
            case HTTP_METHOD_MOVE:
            case HTTP_METHOD_PROPFIND:
            case HTTP_METHOD_PROPPATCH:
            case HTTP_METHOD_LOCK:
            case HTTP_METHOD_UNLOCK:
                break;
            case HTTP_METHOD_OPTIONS:
                /*
                 * 400 is coming from the request-parser BEFORE uri.path is set
                 * 403 is from the response handler when noone else catched it
                 *
                 * */
                if ( (!con->http_status || con->http_status == 200) && con->uri.path->used &&
                     con->uri.path->ptr[0] != '*' )
                {
                    response_header_insert(srv, con, CONST_STR_LEN("Allow"), CONST_STR_LEN("OPTIONS, GET, HEAD, POST"));

                    con->response.transfer_encoding &= ~HTTP_TRANSFER_ENCODING_CHUNKED;
                    con->parsed_response &= ~HTTP_CONTENT_LENGTH;

                    con->http_status = 200;
                    con->file_finished = 1;

                    chunkqueue_reset(con->write_queue);
                }
                break;
            default:
                switch ( con->http_status )
                {
                    case 400: /* bad request */
                    case 414: /* overload request header */
                    case 505: /* unknown protocol */
                    case 207: /* this was webdav */
                        break;
                    default:
                        con->http_status = 501;
                        break;
                }
                break;
        }
    }

    if ( con->http_status == 0 )
    {
        con->http_status = 403;
    }

    switch ( con->http_status )
    {
        case 204: /* class: header only */
        case 205:
        case 304:
            /* disable chunked encoding again as we have no body */
            con->response.transfer_encoding &= ~HTTP_TRANSFER_ENCODING_CHUNKED;
            con->parsed_response &= ~HTTP_CONTENT_LENGTH;
            chunkqueue_reset(con->write_queue);

            con->file_finished = 1;
            break;
#if 0
            /* *** Add upload process, from GXV3175 lighttpd codes, by vanevan *** */
        case 404:
            log_error_write(srv, __FILE__, __LINE__, "ss", "Real path :",
                            con->physical.rel_path->ptr);
            if ( strcasecmp(con->physical.rel_path->ptr, "/preupload") == 0 ||
                 strcasecmp(con->physical.rel_path->ptr, "/upload") == 0 )
            {
                int i = 0;
                struct message m;
                int is_upload = 0;
                log_error_write(srv, __FILE__, __LINE__, "sd",
                                "CONTENT'S LENGTH IS", con->request.content_length);

                if ( strcasecmp(con->physical.rel_path->ptr, "/preupload") == 0 )
                {
                    is_upload = 0;
                }
                else
                {
                    is_upload = 1;
                }

                con->http_status = 200;
                if ( con->mode != DIRECT )
                    break;

                con->file_finished = 0;

                m.hdrcount = 0;
                for ( i = 0; i < MAX_MANHEADERS; i++ )
                {
                    m.headers[0] = NULL;
                }

                //printf("before judeg query\n");
                if ( con->uri.query->ptr != NULL )
                {
                    m.headers[m.hdrcount++] = strtok(con->uri.query->ptr, "&");
                    //printf("con->uri.query->ptr is %s\n", con->uri.query->ptr);
                    while ( m.hdrcount < (MAX_MANHEADERS - 0) &&
                            (m.headers[m.hdrcount] = strtok(NULL, "&")) )
                    {
                        //printf("con->uri.query->ptr is %s\n", con->uri.query->ptr);
                        m.hdrcount++;
                    }
                }

                buffer_reset(con->physical.path);

                if ( !con->file_finished )
                {
                    buffer *b = NULL;

                    con->file_finished = 1;
                    b = chunkqueue_get_append_buffer(con->write_queue);
                    response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/plain"));
                    response_header_overwrite(srv, con, CONST_STR_LEN("Cache-Control"), CONST_STR_LEN("no-cache"));
                    response_header_overwrite(srv, con, CONST_STR_LEN("Pragma"), CONST_STR_LEN("no-cache"));
                    response_header_overwrite(srv, con, CONST_STR_LEN("Access-Control-Allow-Origin"), CONST_STR_LEN("*"));
                    if ( is_upload != 0 )
                    {
                        process_upload(srv, con, b, &m);
                        sync();
                    }
                    else
                    {
                        prepare_upload(srv, con, b, &m);
                    }
                }
                /* Fall through */
                break;
            }
            else if ( strcasecmp(con->physical.rel_path->ptr, "/delete") == 0 )
            {
                log_error_write(srv, __FILE__, __LINE__, "s", "Delete file >>>>>>>>> enter");
                int i = 0;
                int result = 0;
                struct message m;
                char filename[256] = {0}; 
                const char *err = NULL;

                con->http_status = 200;
                if ( con->mode != DIRECT )
                    break;

                m.hdrcount = 0;
                for ( i = 0; i < MAX_MANHEADERS; i++ )
                {
                    m.headers[0] = NULL;
                }

                if ( con->uri.query->ptr != NULL )
                {
                    m.headers[m.hdrcount++] = strtok(con->uri.query->ptr, "&");
                    while ( m.hdrcount < (MAX_MANHEADERS - 0) &&
                            (m.headers[m.hdrcount] = strtok(NULL, "&")) )
                    {
                        m.hdrcount++;
                    }
                }

                buffer *b = NULL;
                b = chunkqueue_get_append_buffer(con->write_queue);
                response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/plain"));

                const char*type = msg_get_header(&m, "type");  /* Delete type */
                const char*data = msg_get_header(&m, "data");  /* Delete data */
                if ( type != NULL && strcasecmp(type, UPLOAD_TYPE_VOICE_PACKEAGE) == 0 && data != NULL )
                {
                    /*Delete Voice package*/
                    result = check_remove_data_from_json(srv, data);
                    if ( result == 0 )
                    {
                        char path[256] = {0};
                        snprintf(path, sizeof(path), "%s%s/%s", "rm -rf ", UPLOAD_SOUNDS_PATH, data);
                        log_error_write(srv, __FILE__, __LINE__, "ss", "Going to remove sounds package: ", path);
                        result = system(path);
                        if ( result != 0 )
                        {
                            err = "Failed to remove language dir!!";
                        }
                    }
                    else
                    {
                        err = "Failed to delete this language!!";
                    }
                }
                else
                {
                    if ( generate_file_name(srv, con, b, &m, filename, sizeof(filename)) != 0 )
                    {
                        log_error_write(srv, __FILE__, __LINE__, "s", "Generate file failed!\n");
                        return -1;
                    }

                    log_error_write(srv, __FILE__, __LINE__, "ss", "file path is ", filename);
                    char rm_cmd[256] = {0};
                    snprintf(rm_cmd, sizeof(rm_cmd), "%s %s", "rm -af", filename);
                    result = system(rm_cmd);
                    log_error_write(srv, __FILE__, __LINE__, "ss", "remove cmd is ", rm_cmd);
                }
                                
                char result_buf[256] = {0};

                if ( result == 0 )
                {
                    sync();
                    snprintf(result_buf, sizeof(result_buf), "%s", "Result=0\r\n");
                }
                else
                {
                    log_error_write(srv, __FILE__, __LINE__, "ss", "remove file error: ", err);
                    snprintf(result_buf, sizeof(result_buf), "%s%s%s", "Result=1\r\nError=", err, "\r\n");
                }
                int len_b = strlen(result_buf);
                char len_result[64] = {0};
                snprintf(len_result, sizeof(len_result), "%ld", len_b);
                response_header_overwrite(srv, con, CONST_STR_LEN("Content-Length"), CONST_STR_LEN(len_result));
                buffer_append_string(b, result_buf);

                log_error_write(srv, __FILE__, __LINE__, "sss", "Delete file ", filename, " >>>>>>>>> finish");

                break;
            }
            else if ( strcasecmp(con->physical.rel_path->ptr, "/info") == 0 )
            {
                log_error_write(srv, __FILE__, __LINE__, "s", "Get Info >>>>>>>>> enter");
                int i = 0;
                struct message m;

                con->http_status = 200;
                if ( con->mode != DIRECT )
                    break;

                m.hdrcount = 0;
                for ( i = 0; i < MAX_MANHEADERS; i++ )
                {
                    m.headers[0] = NULL;
                }

                if ( con->uri.query->ptr != NULL )
                {
                    m.headers[m.hdrcount++] = strtok(con->uri.query->ptr, "&");
                    while ( m.hdrcount < (MAX_MANHEADERS - 0) &&
                            (m.headers[m.hdrcount] = strtok(NULL, "&")) )
                    {
                        m.hdrcount++;
                    }
                }

                buffer *b = NULL;
                b = chunkqueue_get_append_buffer(con->write_queue);
                response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/json;charset=UTF-8"));

                const char*type = msg_get_header(&m, "type");  /* Delete type */
                if ( type != NULL && strcasecmp(type, INFO_TYPE_MODEL_INFO) == 0 )
                {
                    char tmp[256] = {0};
                    char len_buf[64] = {0};
                    const char *name = NULL;
                    const char *description = NULL;
#ifdef BUILD_ON_ARM
                    name = nvram_get(":syslog_id");
                    description = nvram_get("description");
#endif                        
                    snprintf(tmp, sizeof(tmp), "{\"responce\" : \"success\", \"model_name\" : \"%s\", \"description\":\"%s\"}",
                             name == NULL ? "GXE 5120" : name, description == NULL ? "Enterprises PBX" : description);
                    long int len = strlen(tmp);
                    snprintf(len_buf, sizeof(len_buf), "%ld", len);
                    response_header_overwrite(srv, con, CONST_STR_LEN("Content-Length"), CONST_STR_LEN(len_buf));
                    buffer_append_string(b, tmp);
                }

                log_error_write(srv, __FILE__, __LINE__, "s", "Info >>>>>>>>> finish");

                break;
            }
            /*** Add end ***/
#endif

        default: /* class: header + body */
            if ( con->mode != DIRECT ) break;

            /* only custom body for 4xx and 5xx */
            if ( con->http_status < 400 || con->http_status >= 600 ) break;

            con->file_finished = 0;

            buffer_reset(con->physical.path);

            /* try to send static errorfile */
            if ( !buffer_is_empty(con->conf.errorfile_prefix) )
            {
                stat_cache_entry *sce = NULL;

                buffer_copy_string_buffer(con->physical.path, con->conf.errorfile_prefix);
                buffer_append_long(con->physical.path, con->http_status);
                buffer_append_string_len(con->physical.path, CONST_STR_LEN(".html"));

                if ( HANDLER_ERROR != stat_cache_get_entry(srv, con, con->physical.path, &sce) )
                {
                    con->file_finished = 1;

                    http_chunk_append_file(srv, con, con->physical.path, 0, sce->st.st_size);
                    response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_BUF_LEN(sce->content_type));
                }
            }

            if ( !con->file_finished )
            {
                buffer *b;

                buffer_reset(con->physical.path);

                con->file_finished = 1;
                b = chunkqueue_get_append_buffer(con->write_queue);

                /* build default error-page */
                buffer_copy_string_len(b, CONST_STR_LEN(
                                                       "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n"
                                                       "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\"\n"
                                                       "         \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n"
                                                       "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n"
                                                       " <head>\n"
                                                       "  <title>"));
                buffer_append_long(b, con->http_status);
                buffer_append_string_len(b, CONST_STR_LEN(" - "));
                buffer_append_string(b, get_http_status_name(con->http_status));

                buffer_append_string_len(b, CONST_STR_LEN(
                                                         "</title>\n"
                                                         " </head>\n"
                                                         " <body>\n"
                                                         "  <h1>"));
                buffer_append_long(b, con->http_status);
                buffer_append_string_len(b, CONST_STR_LEN(" - "));
                buffer_append_string(b, get_http_status_name(con->http_status));

                buffer_append_string_len(b, CONST_STR_LEN("</h1>\n"
                                                          " </body>\n"
                                                          "</html>\n"
                                                         ));

                response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/html"));
            }
            break;
    }

    if ( con->file_finished )
    {
        /* we have all the content and chunked encoding is not used, set a content-length */

        if ( (!(con->parsed_response & HTTP_CONTENT_LENGTH)) &&
             (con->response.transfer_encoding & HTTP_TRANSFER_ENCODING_CHUNKED) == 0 )
        {
            off_t qlen = chunkqueue_length(con->write_queue);

            /**
             * The Content-Length header only can be sent if we have content:
             * - HEAD doesn't have a content-body (but have a content-length)
             * - 1xx, 204 and 304 don't have a content-body (RFC 2616 Section 4.3)
             *
             * Otherwise generate a Content-Length header as chunked encoding is not 
             * available
             */
            if ( (con->http_status >= 100 && con->http_status < 200) ||
                 con->http_status == 204 ||
                 con->http_status == 304 )
            {
                data_string *ds;
                /* no Content-Body, no Content-Length */
                if ( NULL != (ds = (data_string*) array_get_element(con->response.headers, "Content-Length")) )
                {
                    buffer_reset(ds->value); /* Headers with empty values are ignored for output */
                }
            }
            else if ( qlen > 0 || con->request.http_method != HTTP_METHOD_HEAD )
            {
                /* qlen = 0 is important for Redirects (301, ...) as they MAY have
                 * a content. Browsers are waiting for a Content otherwise
                 */
                buffer_copy_off_t(srv->tmp_buf, qlen);

                response_header_overwrite(srv, con, CONST_STR_LEN("Content-Length"), CONST_BUF_LEN(srv->tmp_buf));
            }
        }
    }
    else
    {
        /**
         * the file isn't finished yet, but we have all headers
         *
         * to get keep-alive we either need:
         * - Content-Length: ... (HTTP/1.0 and HTTP/1.0) or
         * - Transfer-Encoding: chunked (HTTP/1.1)
         */

        if ( ((con->parsed_response & HTTP_CONTENT_LENGTH) == 0) &&
             ((con->response.transfer_encoding & HTTP_TRANSFER_ENCODING_CHUNKED) == 0) )
        {
            con->keep_alive = 0;
        }

        /**
         * if the backend sent a Connection: close, follow the wish
         *
         * NOTE: if the backend sent Connection: Keep-Alive, but no Content-Length, we
         * will close the connection. That's fine. We can always decide the close 
         * the connection
         *
         * FIXME: to be nice we should remove the Connection: ... 
         */
        if ( con->parsed_response & HTTP_CONNECTION )
        {
            /* a subrequest disable keep-alive although the client wanted it */
            if ( con->keep_alive && !con->response.keep_alive )
            {
                con->keep_alive = 0;
            }
        }
    }

    if ( con->request.http_method == HTTP_METHOD_HEAD )
    {
        /**
         * a HEAD request has the same as a GET 
         * without the content
         */
        con->file_finished = 1;

        chunkqueue_reset(con->write_queue);
        con->response.transfer_encoding &= ~HTTP_TRANSFER_ENCODING_CHUNKED;
    }

    http_response_write_header(srv, con);

    return 0;
}

static int connection_handle_write(server *srv, connection *con)
{
    switch ( network_write_chunkqueue(srv, con, con->write_queue) )
    {
        case 0:
            if ( con->file_finished )
            {
                connection_set_state(srv, con, CON_STATE_RESPONSE_END);
                joblist_append(srv, con);
            }
            break;
        case -1: /* error on our side */
            log_error_write(srv, __FILE__, __LINE__, "sd",
                            "connection closed: write failed on fd", con->fd);
            connection_set_state(srv, con, CON_STATE_ERROR);
            joblist_append(srv, con);
            break;
        case -2: /* remote close */
            connection_set_state(srv, con, CON_STATE_ERROR);
            joblist_append(srv, con);
            break;
        case 1:
            con->is_writable = 0;

            /* not finished yet -> WRITE */
            break;
    }

    return 0;
}



connection *connection_init(server *srv)
{
    connection *con;

    UNUSED(srv);

    con = calloc(1, sizeof(*con));

    con->fd = 0;
    con->ndx = -1;
    con->fde_ndx = -1;
    con->bytes_written = 0;
    con->bytes_read = 0;
    con->bytes_header = 0;
    con->loops_per_request = 0;

#define CLEAN(x) \
	con->x = buffer_init();

    CLEAN(request.uri);
    CLEAN(request.request_line);
    CLEAN(request.request);
    CLEAN(request.pathinfo);

    CLEAN(request.orig_uri);

    CLEAN(uri.scheme);
    CLEAN(uri.authority);
    CLEAN(uri.path);
    CLEAN(uri.path_raw);
    CLEAN(uri.query);

    CLEAN(physical.doc_root);
    CLEAN(physical.path);
    CLEAN(physical.basedir);
    CLEAN(physical.rel_path);
    CLEAN(physical.etag);
    CLEAN(parse_request);

    CLEAN(authed_user);
    CLEAN(server_name);
    CLEAN(error_handler);
    CLEAN(dst_addr_buf);
#if defined USE_OPENSSL && ! defined OPENSSL_NO_TLSEXT
    CLEAN(tlsext_server_name);
#endif

#undef CLEAN
    con->write_queue = chunkqueue_init();
    con->read_queue = chunkqueue_init();
    con->request_content_queue = chunkqueue_init();
    chunkqueue_set_tempdirs(con->request_content_queue, srv->srvconf.upload_tempdirs);

    con->request.headers      = array_init();
    con->response.headers     = array_init();
    con->environment     = array_init();

    /* init plugin specific connection structures */

    con->plugin_ctx = calloc(1, (srv->plugins.used + 1) * sizeof(void *));

    con->cond_cache = calloc(srv->config_context->used, sizeof(cond_cache_t));
    config_setup_connection(srv, con);

    return con;
}

void connections_free(server *srv)
{
    connections *conns = srv->conns;
    size_t i;

    for ( i = 0; i < conns->size; i++ )
    {
        connection *con = conns->ptr[i];

        connection_reset(srv, con);

        chunkqueue_free(con->write_queue);
        chunkqueue_free(con->read_queue);
        chunkqueue_free(con->request_content_queue);
        array_free(con->request.headers);
        array_free(con->response.headers);
        array_free(con->environment);

#define CLEAN(x) \
	buffer_free(con->x);

        CLEAN(request.uri);
        CLEAN(request.request_line);
        CLEAN(request.request);
        CLEAN(request.pathinfo);

        CLEAN(request.orig_uri);

        CLEAN(uri.scheme);
        CLEAN(uri.authority);
        CLEAN(uri.path);
        CLEAN(uri.path_raw);
        CLEAN(uri.query);

        CLEAN(physical.doc_root);
        CLEAN(physical.path);
        CLEAN(physical.basedir);
        CLEAN(physical.etag);
        CLEAN(physical.rel_path);
        CLEAN(parse_request);

        CLEAN(authed_user);
        CLEAN(server_name);
        CLEAN(error_handler);
        CLEAN(dst_addr_buf);
#if defined USE_OPENSSL && ! defined OPENSSL_NO_TLSEXT
        CLEAN(tlsext_server_name);
#endif
#undef CLEAN
        free(con->plugin_ctx);
        free(con->cond_cache);

        free(con);
    }

    free(conns->ptr);
}


int connection_reset(server *srv, connection *con)
{
    size_t i;

    plugins_call_connection_reset(srv, con);

    con->is_readable = 1;
    con->is_writable = 1;
    con->http_status = 0;
    con->file_finished = 0;
    con->file_started = 0;
    con->got_response = 0;

    con->parsed_response = 0;

    con->bytes_written = 0;
    con->bytes_written_cur_second = 0;
    con->bytes_read = 0;
    con->bytes_header = 0;
    con->loops_per_request = 0;

    con->request.http_method = HTTP_METHOD_UNSET;
    con->request.http_version = HTTP_VERSION_UNSET;

    con->request.http_if_modified_since = NULL;
    con->request.http_if_none_match = NULL;

    con->response.keep_alive = 0;
    con->response.content_length = -1;
    con->response.transfer_encoding = 0;

    con->mode = DIRECT;

#define CLEAN(x) \
	if (con->x) buffer_reset(con->x);

    CLEAN(request.uri);
    CLEAN(request.request_line);
    CLEAN(request.pathinfo);
    CLEAN(request.request);

    /* CLEAN(request.orig_uri); */

    CLEAN(uri.scheme);
    /* CLEAN(uri.authority); */
    /* CLEAN(uri.path); */
    CLEAN(uri.path_raw);
    /* CLEAN(uri.query); */

    CLEAN(physical.doc_root);
    CLEAN(physical.path);
    CLEAN(physical.basedir);
    CLEAN(physical.rel_path);
    CLEAN(physical.etag);

    CLEAN(parse_request);

    CLEAN(authed_user);
    CLEAN(server_name);
    CLEAN(error_handler);
#if defined USE_OPENSSL && ! defined OPENSSL_NO_TLSEXT
    CLEAN(tlsext_server_name);
#endif
#undef CLEAN

#define CLEAN(x) \
	if (con->x) con->x->used = 0;

#undef CLEAN

#define CLEAN(x) \
		con->request.x = NULL;

    CLEAN(http_host);
    CLEAN(http_range);
    CLEAN(http_content_type);
#undef CLEAN
    con->request.content_length = 0;

    array_reset(con->request.headers);
    array_reset(con->response.headers);
    array_reset(con->environment);

    chunkqueue_reset(con->write_queue);
    chunkqueue_reset(con->request_content_queue);

    /* the plugins should cleanup themself */
    for ( i = 0; i < srv->plugins.used; i++ )
    {
        plugin *p = ((plugin **)(srv->plugins.ptr))[i];
        plugin_data *pd = p->data;

        if ( !pd ) continue;

        if ( con->plugin_ctx[pd->id] != NULL )
        {
            log_error_write(srv, __FILE__, __LINE__, "sb", "missing cleanup in", p->name);
        }

        con->plugin_ctx[pd->id] = NULL;
    }

    /* The cond_cache gets reset in response.c */
    /* config_cond_cache_reset(srv, con); */

    con->header_len = 0;
    con->in_error_handler = 0;

    config_setup_connection(srv, con);

    return 0;
}

#if 0
/*Added Process for Voice package, rzhang*/
/*Need to free the return string*/
static char *process_voice_package(server *srv, const char *filename)
{
    log_error_write(srv, __FILE__, __LINE__, "ss", "process voice package !!!!!!!", filename);

    char *ret = NULL;
    int _index = 0;
    int total = 0;
    char cmd[256] = {0};
    char buf[256] = {0};
    char result[1024] = {0};
    FILE *output = NULL;
    char lang_id[8] = {0};
    char lang_name[256] = {0};

    struct TarCMDS
    {
        const char *suffix;
        const char *cmd;
    };

    /*.tar     => tar -xvf 
      .gz      => gunzip
      .tar.gz  => tar -xzf
      .tgz     => tar -xzf
      .bz2     => bunzip2
      .tar.bz2 => tar -xjf
      .Z       => uncompress(unspported)
      .tar.Z   => tar -xZf
      .rar     => unrar(unspported) 
      .zip     => unzip*/

    struct TarCMDS cmds[] = {
        { ".tar.bz2" , "tar -xjf" },
        { ".tar.gz"  , "tar -xzf" },
        { ".tar.Z"   , "tar -xZf" },
        { ".tgz"     , "tar -xzf" },
        { ".tar"     , "tar -xf"  },
        { ".bz2"     , "bunzip2"  },
        { ".zip"     , "unzip -q" },
        { ".gz"      , "gunzip"   }
    };

    if ( filename == NULL )
    {
        snprintf(result, sizeof(result), "Filename is NULL!");
        log_error_write(srv, __FILE__, __LINE__, "s", result);

        goto err;
    }
    total = sizeof(cmds) / sizeof(struct TarCMDS);
    log_error_write(srv, __FILE__, __LINE__, "sd", "TAR CMDS total is ", total);
    
    int err = chdir( UPLOAD_SOUNDS_TMP_PATH );
    
    if ( err < 0 )
    {
        char * mesg = strerror(errno);  
        log_error_write(srv, __FILE__, __LINE__, "ss", "chdir fail, msg is", mesg);
    }
    
    while ( _index < total )
    {
        if ( strcasestr(filename, cmds[_index].suffix ) != NULL )
        {
            snprintf(cmd, sizeof(cmd), "%s %s", cmds[_index].cmd, filename);
            log_error_write(srv, __FILE__, __LINE__, "ss", "Untar cmd is ", cmd);

            output = popen(cmd, "r");
            while ( output != NULL && (!feof(output)) && (fgets(buf, sizeof(buf) - 1, output) != NULL) )
            {
                log_error_write(srv, __FILE__, __LINE__, "ss", "Untar output is ", buf);
                strncat(result, buf, (sizeof(result)-1 - strlen(result)));
            }

            /*Judge output*/
            if ( result[0] != '\0' )
            {
                log_error_write(srv, __FILE__, __LINE__, "ss", "Untar fail: ", result);
            }

            if ( output != NULL )
            {
                pclose(output);
                output = NULL;
            }

            //snprintf(cmd, sizeof(cmd), "rm %s", filename);
            //system(cmd);
            break;
        }
        _index++;
    }

    if ( result[0] == '\0' )
    {
        /*Parse the language information txt file*/
        FILE *ptr = fopen(UPLOAD_SOUNDS_TMP_PATH"/info.txt", "r");
        if ( ptr != NULL )
        {
            fseek(ptr, 0, SEEK_SET);//start from the start of file
            fgets(lang_name, sizeof(lang_name)-1, ptr);
            char *sptr = strchr(lang_name, '\n');
            if ( sptr != NULL )
            {
                *sptr = '\0';
            }
            log_error_write(srv, __FILE__, __LINE__, "ssss", "Language name is: ", "[", lang_name, "]");
            fclose(ptr);
            ptr = NULL;
            if ( lang_name[0] == '\0' )
            {
                log_error_write(srv, __FILE__, __LINE__, "s", "Invalid package! No info in the info.txt file!");
                strcat(result, "Invalid package! No info in the info.txt file!");
                goto err;
            }
        }
        else
        {
            /*Invalid package!*/
            log_error_write(srv, __FILE__, __LINE__, "s", "Invalid package! No info.txt file!");
            strcat(result, "Invalid package! No info.txt file!");
            goto err;
        }

        /*Get the dir name of sounds file*/
        DIR *dir = NULL;
        struct dirent *direct = NULL;
        dir = opendir(UPLOAD_SOUNDS_TMP_PATH);
        if ( dir == NULL )
        {
            log_error_write(srv, __FILE__, __LINE__, "ss", "Open DIR failed!", UPLOAD_SOUNDS_TMP_PATH);
            strcat(result, "Open DIR failed!");
            goto err;
        }

        while ( (direct = readdir(dir) ) != NULL )
        {
            if ( direct->d_type == DT_DIR 
                 && strcmp(direct->d_name, ".") != 0
                 && strcmp(direct->d_name, "..") != 0 )
            {
                if ( lang_id[0] == '\0' )
                {
                    strncat(lang_id, direct->d_name, sizeof(lang_id) - 1);
                    log_error_write(srv, __FILE__, __LINE__, "ss", "Language DIR is: ", lang_id);
                }
                else
                {
                    log_error_write(srv, __FILE__, __LINE__, "s", "More than one dir in the package!");

                    strcat(result, "Invalid package! More than one dir in the package!");
                    closedir(dir);
                    dir = NULL;
                    goto err;
                }
            }
        }
        closedir(dir);
        dir = NULL;

        /*Added the new option to language.json*/
        if ( add_data_to_json( srv, lang_id, lang_name ) == 0 )
        {
            /*Success*/
            char mv_cmd[256] = {0};
            snprintf(mv_cmd, sizeof(mv_cmd), "mv %s/%s %s/%s", UPLOAD_SOUNDS_TMP_PATH, lang_id, UPLOAD_SOUNDS_PATH, lang_id);
            int res = system(mv_cmd);
            if ( res != 0 )
            {
                log_error_write(srv, __FILE__, __LINE__, "s", "Failed when installing voice package!");
                strcat(result, "Failed when installing voice package!");
            }
        }
        else
        {
            /*Failed*/
            strcat(result, "Added voice package failed!");
        }
    }

    err:
    remove(filename);
    memset(cmd, 0, sizeof(cmd));
    snprintf(cmd, sizeof(cmd), "%s%s", "rm -rf ", UPLOAD_SOUNDS_TMP_PATH);
    system(cmd);

    if ( _index > total || result[0] != '\0'  )
    {
        ret = strdup(result);
    }

    return ret;
}
/*Added end*/
#endif

/*Added Process for Zero Configuration, rzhang*/
static buffer *process_zero_config(server *srv, buffer *_request)
{
    log_error_write(srv, __FILE__, __LINE__, "s", "process_zero_config !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");

    if ( _request == NULL )
    {
        log_error_write(srv, __FILE__, __LINE__, "s", "request is NULL !!!!!!");
        return NULL;
    }

    int success = 0;
    char mac[16] = {0};
    const char *zccgi = "/zccgi/";
    char *new_uri = NULL;
    char *path = NULL;
    char *req = NULL;
    buffer *tmp = buffer_init_buffer(_request);

    /*Judge the URI*/
    if ( tmp != NULL && tmp->ptr != NULL )
    {
        req = strdup(tmp->ptr);
        if ( req == NULL )
        {
            log_error_write(srv, __FILE__, __LINE__, "s", "Malloc failed!");
            goto failed_return;
        }
        char *first_line_start = req;
        char *first_line_end = strstr(req, "\r\n");
        if ( first_line_end != NULL )
        {
            *first_line_end = '\0';
        }
        else
        {
            log_error_write(srv, __FILE__, __LINE__, "s", "Fail to find the '\r'!");
            goto failed_return;
        }
        char *first_line = first_line_start;
        log_error_write(srv, __FILE__, __LINE__, "ss", "First line is", first_line);

        if ( strstr(first_line, "GET") != NULL && strchr(first_line, '?') == NULL )
        {
            /*Parse the path*/
            char *path_start = strchr(first_line, '/');
            char *path_end = strstr(first_line, "HTTP");
            int path_len = path_end - path_start;
            path = (char*)malloc(path_len);
            if ( path == NULL )
            {
                log_error_write(srv, __FILE__, __LINE__, "s", "Malloc failed!");
                goto failed_return;
            }
            memset(path, 0, path_len);
            memcpy(path, path_start, path_len-1);
            log_error_write(srv, __FILE__, __LINE__, "ss", "PATH is ", path);
            int count = 0;
            char *line_ptr = path;
            while ( (line_ptr = strchr(line_ptr, '/')) != NULL )
            {
                /*only include one '/'*/
                count++;
                line_ptr++;
                if ( count > 1 )
                {
                    log_error_write(srv, __FILE__, __LINE__, "s", "PATH is including more than one '/' !!!");
                    goto failed_return;
                }
            }

            char *tmp_ptr = path;
            while ( tmp_ptr != NULL && *tmp_ptr != '\0' )
            {
                sscanf(tmp_ptr, "%12[0-9a-fA-F]", mac);
                if ( strlen(mac) == 12 )
                {
                    break;
                }
                tmp_ptr++;
            }

            log_error_write(srv, __FILE__, __LINE__, "ss", "MAC:", mac);
            if ( strlen(mac) == 12 && 
                 ( strcasestr(path, ".xml") != NULL || strcasestr(path, ".cfg") != NULL ) )
            {
                log_error_write(srv, __FILE__, __LINE__, "ss", "OK:", "IS xml or cfg format !!!!!!!!!!!!!!!!!!");
                int len = strlen(path) + strlen(zccgi) + 1;
                new_uri = (char *)malloc( len * sizeof(char) );

                if ( new_uri != NULL )
                {
                    memset(new_uri, 0, len);
                    char *ptr = strchr(path, '/');
                    if ( ptr == NULL )
                    {
                        snprintf(new_uri, len, "%s%s", zccgi, path);
                    }
                    else
                    {
                        log_error_write(srv, __FILE__, __LINE__, "ss", "PTR is", ptr);
                        if ( ptr + 1 != NULL )
                        {
                            /*Remove the first '/'*/
                            snprintf( new_uri, len, "%s%s", zccgi, ptr+1 );
                        }
                    }
                    log_error_write(srv, __FILE__, __LINE__, "ss", "new_uri is", new_uri);

                    /*Create the new request buffer*/
                    buffer_reset(tmp);
                    *path_start = '\0';

                    buffer_append_string(tmp, first_line);
                    buffer_append_string(tmp, new_uri);
                    buffer_append_string(tmp, " ");
                    buffer_append_string(tmp, path_end);
                    buffer_append_string(tmp, "\r\n");
                    buffer_append_string(tmp, first_line_end+2);/*2 is strlen "\r\n"*/
                    success = 1;

                    log_error_write(srv, __FILE__, __LINE__, "ss", "new_request is", tmp->ptr);

                }
            }
        }
    }

    failed_return:
    if ( new_uri != NULL )
    {
        free(new_uri);
    }

    if ( path != NULL )
    {
        free(path);
    }

    if ( req != NULL )
    {
        free(req);
    }

    if ( success == 0 )
    {
        buffer_free(tmp);
        tmp = NULL;
    }

    return tmp;
}
/*Added end*/

/**
 * handle all header and content read
 *
 * we get called by the state-engine and by the fdevent-handler
 */
static int connection_handle_read_state(server *srv, connection *con)
{
    connection_state_t ostate = con->state;
    chunk *c, *last_chunk;
    off_t last_offset;
    chunkqueue *cq = con->read_queue;
    chunkqueue *dst_cq = con->request_content_queue;
    int is_closed = 0; /* the connection got closed, if we don't have a complete header, -> error */

    if ( con->is_readable )
    {
        con->read_idle_ts = srv->cur_ts;

        switch ( connection_handle_read(srv, con) )
        {
            case -1:
                return -1;
            case -2:
                is_closed = 1;
                break;
            default:
                break;
        }
    }

    if ( con->request.request != NULL && con->request.request->ptr != NULL 
         && strcasestr(con->request.request->ptr, "/upload?") != NULL )
    {
        /*Process upload progress. Added by rzhang*/
#ifdef BUILD_ON_ARM
        if ( s_setting_nvram == 0 )
        {
            s_setting_nvram++;
            char pvalue[256] = {0};
            snprintf(pvalue, sizeof(pvalue), "%ld", (long int)con->request.content_length);
            nvram_set(UPLOAD_TOTAL, pvalue);
            snprintf(pvalue, sizeof(pvalue), "%ld", (long int)con->bytes_read);
            nvram_set(UPLOAD_RECV, pvalue);
            s_setting_nvram--;
            //nvram_set(UPLOAD_PROGRESS, "0");
        }
#endif
        log_error_write(srv, __FILE__, __LINE__, "sd", "UPLOAD TOTAL SIZE IS:", con->request.content_length);
        //log_error_write(srv, __FILE__, __LINE__, "ss", "URL:", con->request.request->ptr);
        log_error_write(srv, __FILE__, __LINE__, "sO", "UPLOAD SIZE IS:", con->bytes_read);
        /*Added end*/
    }
    else
    {
        log_error_write(srv, __FILE__, __LINE__, "s", "Isn't UPLOAD request !!!!!!!!!!!!!");
    }

    /* the last chunk might be empty */
    for ( c = cq->first; c; )
    {
        if ( cq->first == c && c->mem->used == 0 )
        {
            /* the first node is empty */
            /* ... and it is empty, move it to unused */

            cq->first = c->next;
            if ( cq->first == NULL ) cq->last = NULL;

            c->next = cq->unused;
            cq->unused = c;
            cq->unused_chunks++;

            c = cq->first;
        }
        else if ( c->next && c->next->mem->used == 0 )
        {
            chunk *fc;
            /* next node is the last one */
            /* ... and it is empty, move it to unused */

            fc = c->next;
            c->next = fc->next;

            fc->next = cq->unused;
            cq->unused = fc;
            cq->unused_chunks++;

            /* the last node was empty */
            if ( c->next == NULL )
            {
                cq->last = c;
            }

            c = c->next;
        }
        else
        {
            c = c->next;
        }
    }

    /* we might have got several packets at once
     */

    switch ( ostate )
    {
        case CON_STATE_READ:
            /* if there is a \r\n\r\n in the chunkqueue
             *
             * scan the chunk-queue twice
             * 1. to find the \r\n\r\n
             * 2. to copy the header-packet
             *
             */

            last_chunk = NULL;
            last_offset = 0;

            for ( c = cq->first; c; c = c->next )
            {
                buffer b;
                size_t i;

                b.ptr = c->mem->ptr + c->offset;
                b.used = c->mem->used - c->offset;
                if ( b.used > 0 ) b.used--; /* buffer "used" includes terminating zero */

                for ( i = 0; i < b.used; i++ )
                {
                    char ch = b.ptr[i];

                    if ( '\r' == ch )
                    {
                        /* chec if \n\r\n follows */
                        size_t j = i+1;
                        chunk *cc = c;
                        const char header_end[] = "\r\n\r\n";
                        int header_end_match_pos = 1;

                        for ( ; cc; cc = cc->next, j = 0 )
                        {
                            buffer bb;
                            bb.ptr = cc->mem->ptr + cc->offset;
                            bb.used = cc->mem->used - cc->offset;
                            if ( bb.used > 0 ) bb.used--; /* buffer "used" includes terminating zero */

                            for ( ; j < bb.used; j++ )
                            {
                                ch = bb.ptr[j];

                                if ( ch == header_end[header_end_match_pos] )
                                {
                                    header_end_match_pos++;
                                    if ( 4 == header_end_match_pos )
                                    {
                                        last_chunk = cc;
                                        last_offset = j+1;
                                        goto found_header_end;
                                    }
                                }
                                else
                                {
                                    goto reset_search;
                                }
                            }
                        }
                    }
                    reset_search: ;
                }
            }
            found_header_end:

            /* found */
            if ( last_chunk )
            {
                buffer_reset(con->request.request);

                for ( c = cq->first; c; c = c->next )
                {
                    buffer b;

                    b.ptr = c->mem->ptr + c->offset;
                    b.used = c->mem->used - c->offset;

                    if ( c == last_chunk )
                    {
                        b.used = last_offset + 1;
                    }
                    
                    //log_error_write(srv, __FILE__, __LINE__, "sss", "b1111 is [", b.ptr, "]");
                    buffer *new_req = process_zero_config(srv, &b);
                    if ( new_req != NULL )
                    {
                        buffer_append_string_buffer(con->request.request, new_req );
                        buffer_copy_string_buffer(&b, new_req);
                        buffer_free(new_req);
                        new_req = NULL;
                    }
                    else
                    {
                        buffer_append_string_buffer(con->request.request, &b);
                    }
                    //log_error_write(srv, __FILE__, __LINE__, "sss", "b2222 is [", b.ptr, "]"); 
                     
                    //buffer_append_string_buffer(con->request.request, &b);

                    if ( c == last_chunk )
                    {
                        c->offset += last_offset;

                        break;
                    }
                    else
                    {
                        /* the whole packet was copied */
                        c->offset = c->mem->used - 1;
                    }
                }

                connection_set_state(srv, con, CON_STATE_REQUEST_END);
            }
            else if ( chunkqueue_length(cq) > 64 * 1024 )
            {
                log_error_write(srv, __FILE__, __LINE__, "s", "oversized request-header -> sending Status 414");

                con->http_status = 414; /* Request-URI too large */
                con->keep_alive = 0;
                connection_set_state(srv, con, CON_STATE_HANDLE_REQUEST);
            }
            break;
        case CON_STATE_READ_POST:
            for ( c = cq->first; c && (dst_cq->bytes_in != (off_t)con->request.content_length); c = c->next )
            {
                off_t weWant, weHave, toRead;

                weWant = con->request.content_length - dst_cq->bytes_in;

                assert(c->mem->used);

                weHave = c->mem->used - c->offset - 1;

                toRead = weHave > weWant ? weWant : weHave;

                /* the new way, copy everything into a chunkqueue whcih might use tempfiles */
                if ( con->request.content_length > 64 * 1024 )
                {
                    chunk *dst_c = NULL;
                    /* copy everything to max 1Mb sized tempfiles */

                    /*
                     * if the last chunk is
                     * - smaller than 1Mb (size < 1Mb)
                     * - not read yet (offset == 0)
                     * -> append to it
                     * otherwise
                     * -> create a new chunk
                     *
                     * */

                    if ( dst_cq->last &&
                         dst_cq->last->type == FILE_CHUNK &&
                         dst_cq->last->file.is_temp &&
                         dst_cq->last->offset == 0 )
                    {
                        /* ok, take the last chunk for our job */

                        if ( dst_cq->last->file.length < 1 * 1024 * 1024 )
                        {
                            dst_c = dst_cq->last;

                            if ( dst_c->file.fd == -1 )
                            {
                                /* this should not happen as we cache the fd, but you never know */
                                dst_c->file.fd = open(dst_c->file.name->ptr, O_WRONLY | O_APPEND);
#ifdef FD_CLOEXEC
                                fcntl(dst_c->file.fd, F_SETFD, FD_CLOEXEC);
#endif
                            }
                        }
                        else
                        {
                            /* the chunk is too large now, close it */
                            dst_c = dst_cq->last;

                            if ( dst_c->file.fd != -1 )
                            {
                                close(dst_c->file.fd);
                                dst_c->file.fd = -1;
                            }
                            dst_c = chunkqueue_get_append_tempfile(dst_cq);
                        }
                    }
                    else
                    {
                        dst_c = chunkqueue_get_append_tempfile(dst_cq);
                    }

                    /* we have a chunk, let's write to it */

                    if ( dst_c->file.fd == -1 )
                    {
                        /* we don't have file to write to,
                         * EACCES might be one reason.
                         *
                         * Instead of sending 500 we send 413 and say the request is too large
                         *  */

                        log_error_write(srv, __FILE__, __LINE__, "sbs",
                                        "denying upload as opening to temp-file for upload failed:",
                                        dst_c->file.name, strerror(errno));

                        con->http_status = 413; /* Request-Entity too large */
                        con->keep_alive = 0;
                        connection_set_state(srv, con, CON_STATE_HANDLE_REQUEST);

                        break;
                    }

                    if ( toRead != write(dst_c->file.fd, c->mem->ptr + c->offset, toRead) )
                    {
                        /* write failed for some reason ... disk full ? */
                        log_error_write(srv, __FILE__, __LINE__, "sbs",
                                        "denying upload as writing to file failed:",
                                        dst_c->file.name, strerror(errno));

                        con->http_status = 413; /* Request-Entity too large */
                        con->keep_alive = 0;
                        connection_set_state(srv, con, CON_STATE_HANDLE_REQUEST);

                        close(dst_c->file.fd);
                        dst_c->file.fd = -1;

                        break;
                    }

                    dst_c->file.length += toRead;

                    if ( dst_cq->bytes_in + toRead == (off_t)con->request.content_length )
                    {
                        /* we read everything, close the chunk */
                        close(dst_c->file.fd);
                        dst_c->file.fd = -1;
                    }
                }
                else
                {
                    buffer *b;

                    if ( dst_cq->last &&
                         dst_cq->last->type == MEM_CHUNK )
                    {
                        b = dst_cq->last->mem;
                    }
                    else
                    {
                        b = chunkqueue_get_append_buffer(dst_cq);
                        /* prepare buffer size for remaining POST data; is < 64kb */
                        buffer_prepare_copy(b, con->request.content_length - dst_cq->bytes_in + 1);
                    }
                    buffer_append_string_len(b, c->mem->ptr + c->offset, toRead);
                }

                c->offset += toRead;
                dst_cq->bytes_in += toRead;
            }

            /* Content is ready */
            if ( dst_cq->bytes_in == (off_t)con->request.content_length )
            {
                connection_set_state(srv, con, CON_STATE_HANDLE_REQUEST);
            }

            break;
        default: break;
    }

    /* the connection got closed and we didn't got enough data to leave one of the READ states
     * the only way is to leave here */
    if ( is_closed && ostate == con->state )
    {
        connection_set_state(srv, con, CON_STATE_ERROR);
    }

    chunkqueue_remove_finished_chunks(cq);

    return 0;
}

static handler_t connection_handle_fdevent(server *srv, void *context, int revents)
{
    connection *con = context;

    joblist_append(srv, con);

    if ( con->conf.is_ssl )
    {
        /* ssl may read and write for both reads and writes */
        if ( revents & (FDEVENT_IN | FDEVENT_OUT) )
        {
            con->is_readable = 1;
            con->is_writable = 1;
        }
    }
    else
    {
        if ( revents & FDEVENT_IN )
        {
            con->is_readable = 1;
        }
        if ( revents & FDEVENT_OUT )
        {
            con->is_writable = 1;
            /* we don't need the event twice */
        }
    }


    if ( revents & ~(FDEVENT_IN | FDEVENT_OUT) )
    {
        /* looks like an error */

        /* FIXME: revents = 0x19 still means that we should read from the queue */
        if ( revents & FDEVENT_HUP )
        {
            if ( con->state == CON_STATE_CLOSE )
            {
                con->close_timeout_ts = srv->cur_ts - (HTTP_LINGER_TIMEOUT+1);
            }
            else
            {
                /* sigio reports the wrong event here
                 *
                 * there was no HUP at all
                 */
#ifdef USE_LINUX_SIGIO
                if ( srv->ev->in_sigio == 1 )
                {
                    log_error_write(srv, __FILE__, __LINE__, "sd",
                                    "connection closed: poll() -> HUP", con->fd);
                }
                else
                {
                    connection_set_state(srv, con, CON_STATE_ERROR);
                }
#else
                connection_set_state(srv, con, CON_STATE_ERROR);
#endif

            }
        }
        else if ( revents & FDEVENT_ERR )
        {
            /* error, connection reset, whatever... we don't want to spam the logfile */
#if 0
            log_error_write(srv, __FILE__, __LINE__, "sd",
                            "connection closed: poll() -> ERR", con->fd);
#endif
            connection_set_state(srv, con, CON_STATE_ERROR);
        }
        else
        {
            log_error_write(srv, __FILE__, __LINE__, "sd",
                            "connection closed: poll() -> ???", revents);
        }
    }

    if ( con->state == CON_STATE_READ ||
         con->state == CON_STATE_READ_POST )
    {
        connection_handle_read_state(srv, con);
    }

    if ( con->state == CON_STATE_WRITE &&
         !chunkqueue_is_empty(con->write_queue) &&
         con->is_writable )
    {

        if ( -1 == connection_handle_write(srv, con) )
        {
            connection_set_state(srv, con, CON_STATE_ERROR);

            log_error_write(srv, __FILE__, __LINE__, "ds",
                            con->fd,
                            "handle write failed.");
        }
        else if ( con->state == CON_STATE_WRITE )
        {
            con->write_request_ts = srv->cur_ts;
        }
    }

    if ( con->state == CON_STATE_CLOSE )
    {
        /* flush the read buffers */
        int len;
        char buf[1024];

        len = read(con->fd, buf, sizeof(buf));
        if ( len == 0 || (len < 0 && errno != EAGAIN && errno != EINTR) )
        {
            con->close_timeout_ts = srv->cur_ts - (HTTP_LINGER_TIMEOUT+1);
        }
    }

    return HANDLER_FINISHED;
}


connection *connection_accept(server *srv, server_socket *srv_socket)
{
    /* accept everything */

    /* search an empty place */
    int cnt;
    sock_addr cnt_addr;
    socklen_t cnt_len;
    /* accept it and register the fd */

    /**
     * check if we can still open a new connections
     *
     * see #1216
     */

    if ( srv->conns->used >= srv->max_conns )
    {
        return NULL;
    }

    cnt_len = sizeof(cnt_addr);

    if ( -1 == (cnt = accept(srv_socket->fd, (struct sockaddr *) &cnt_addr, &cnt_len)) )
    {
        switch ( errno )
        {
            case EAGAIN:
#if EWOULDBLOCK != EAGAIN
            case EWOULDBLOCK:
#endif
            case EINTR:
                /* we were stopped _before_ we had a connection */
            case ECONNABORTED: /* this is a FreeBSD thingy */
                /* we were stopped _after_ we had a connection */
                break;
            case EMFILE:
                /* out of fds */
                break;
            default:
                log_error_write(srv, __FILE__, __LINE__, "ssd", "accept failed:", strerror(errno), errno);
        }
        return NULL;
    }
    else
    {
        connection *con;

        srv->cur_fds++;

        /* ok, we have the connection, register it */
#if 0
        log_error_write(srv, __FILE__, __LINE__, "sd",
                        "appected()", cnt);
#endif
        srv->con_opened++;

        con = connections_get_new_connection(srv);

        con->fd = cnt;
        con->fde_ndx = -1;
#if 0
        gettimeofday(&(con->start_tv), NULL);
#endif
        fdevent_register(srv->ev, con->fd, connection_handle_fdevent, con);

        connection_set_state(srv, con, CON_STATE_REQUEST_START);

        con->connection_start = srv->cur_ts;
        con->dst_addr = cnt_addr;
        buffer_copy_string(con->dst_addr_buf, inet_ntop_cache_get_ip(srv, &(con->dst_addr)));
        con->srv_socket = srv_socket;

        if ( -1 == (fdevent_fcntl_set(srv->ev, con->fd)) )
        {
            log_error_write(srv, __FILE__, __LINE__, "ss", "fcntl failed: ", strerror(errno));
            return NULL;
        }
#ifdef USE_OPENSSL
        /* connect FD to SSL */
        if ( srv_socket->is_ssl )
        {
            if ( NULL == (con->ssl = SSL_new(srv_socket->ssl_ctx)) )
            {
                log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:",
                                ERR_error_string(ERR_get_error(), NULL));

                return NULL;
            }

#ifndef OPENSSL_NO_TLSEXT
            SSL_set_app_data(con->ssl, con);
#endif
            SSL_set_accept_state(con->ssl);
            con->conf.is_ssl=1;

            if ( 1 != (SSL_set_fd(con->ssl, cnt)) )
            {
                log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:",
                                ERR_error_string(ERR_get_error(), NULL));
                return NULL;
            }
        }
#endif
        return con;
    }
}

int connection_state_machine(server *srv, connection *con)
{
   // log_error_write(srv, __FILE__, __LINE__, "sss[d]", ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>",
   //                 con->uri.path->ptr,
   //                 con->uri.query->ptr, con->state);


    if ( con->state == CON_STATE_READ_POST && con->request.request != NULL && con->request.request->ptr != NULL && strcasestr(con->request.request->ptr, "/upload?") != NULL )
    {
        //int progress = (100*con->bytes_read)/con->request.content_length;

#ifdef BUILD_ON_ARM
        if ( s_setting_nvram == 0 )
        {
            s_setting_nvram++;
            char pvalue[128] = {0};
            //char progress_buf[8] = {0};
            snprintf(pvalue, sizeof(pvalue), "%ld", (long int)con->bytes_read);
            int restluuu = nvram_set(UPLOAD_RECV, pvalue);
            if ( restluuu == 0 )
            {
                log_error_write(srv, __FILE__, __LINE__, "sssd", "buffer received: ", pvalue, "buffer total: ", con->request.content_length);
            }
            else
            {
                log_error_write(srv, __FILE__, __LINE__, "d sssd", restluuu, "buffer received: ", pvalue, "buffer total: ", con->request.content_length);
            }
            s_setting_nvram--;
        }
        else
        {
            log_error_write(srv, __FILE__, __LINE__, "s", "NVRAM IS SETTING!!!!!!!!!!!!");
        }

        //snprintf(progress_buf, sizeof(progress_buf), "%d", progress);
        //nvram_set(UPLOAD_PROGRESS, progress_buf);
        //log_error_write(srv, __FILE__, __LINE__, "ss", "progress is: ", progress_buf );
#endif
    }

    int done = 0, r;
#ifdef USE_OPENSSL
    server_socket *srv_sock = con->srv_socket;
#endif

    if ( srv->srvconf.log_state_handling )
    {
        log_error_write(srv, __FILE__, __LINE__, "sds",
                        "state at start",
                        con->fd,
                        connection_get_state(con->state));
    }

    while ( done == 0 )
    {
        size_t ostate = con->state;

        switch ( con->state )
        {
            case CON_STATE_REQUEST_START: /* transient */
                if ( srv->srvconf.log_state_handling )
                {
                    log_error_write(srv, __FILE__, __LINE__, "sds",
                                    "state for fd", con->fd, connection_get_state(con->state));
                }

                con->request_start = srv->cur_ts;
                con->read_idle_ts = srv->cur_ts;

                con->request_count++;
                con->loops_per_request = 0;

                connection_set_state(srv, con, CON_STATE_READ);

                /* patch con->conf.is_ssl if the connection is a ssl-socket already */

#ifdef USE_OPENSSL
                con->conf.is_ssl = srv_sock->is_ssl;
#endif

                break;
            case CON_STATE_REQUEST_END: /* transient */
                if ( srv->srvconf.log_state_handling )
                {
                    log_error_write(srv, __FILE__, __LINE__, "sds",
                                    "state for fd", con->fd, connection_get_state(con->state));
                }

                buffer_reset(con->uri.authority);
                buffer_reset(con->uri.path);
                buffer_reset(con->uri.query);
                buffer_reset(con->request.orig_uri);

                if ( http_request_parse(srv, con) )
                {
                    /* we have to read some data from the POST request */

                    connection_set_state(srv, con, CON_STATE_READ_POST);

                    break;
                }

                connection_set_state(srv, con, CON_STATE_HANDLE_REQUEST);

                break;
            case CON_STATE_HANDLE_REQUEST:
                /*
                 * the request is parsed
                 *
                 * decided what to do with the request
                 * -
                 *
                 *
                 */

                if ( srv->srvconf.log_state_handling )
                {
                    log_error_write(srv, __FILE__, __LINE__, "sds",
                                    "state for fd", con->fd, connection_get_state(con->state));
                }

                switch ( r = http_response_prepare(srv, con) )
                {
                    case HANDLER_FINISHED:
                        if ( con->mode == DIRECT )
                        {
                            if ( con->http_status == 404 ||
                                 con->http_status == 403 )
                            {
                                /* 404 error-handler */

                                if ( con->in_error_handler == 0 &&
                                     (!buffer_is_empty(con->conf.error_handler) ||
                                      !buffer_is_empty(con->error_handler)) )
                                {
                                    /* call error-handler */

                                    con->error_handler_saved_status = con->http_status;
                                    con->http_status = 0;

                                    if ( buffer_is_empty(con->error_handler) )
                                    {
                                        buffer_copy_string_buffer(con->request.uri, con->conf.error_handler);
                                    }
                                    else
                                    {
                                        buffer_copy_string_buffer(con->request.uri, con->error_handler);
                                    }
                                    buffer_reset(con->physical.path);

                                    con->in_error_handler = 1;

                                    connection_set_state(srv, con, CON_STATE_HANDLE_REQUEST);

                                    done = -1;
                                    break;
                                }
                                else if ( con->in_error_handler )
                                {
                                    /* error-handler is a 404 */

                                    con->http_status = con->error_handler_saved_status;
                                }
                            }
                            else if ( con->in_error_handler )
                            {
                                /* error-handler is back and has generated content */
                                /* if Status: was set, take it otherwise use 200 */
                            }
                        }
                        if ( con->http_status == 0 ) con->http_status = 200;

                        /* we have something to send, go on */
                        connection_set_state(srv, con, CON_STATE_RESPONSE_START);
                        break;
                    case HANDLER_WAIT_FOR_FD:
                        srv->want_fds++;

                        fdwaitqueue_append(srv, con);

                        connection_set_state(srv, con, CON_STATE_HANDLE_REQUEST);

                        break;
                    case HANDLER_COMEBACK:
                        done = -1;
                    case HANDLER_WAIT_FOR_EVENT:
                        /* come back here */
                        connection_set_state(srv, con, CON_STATE_HANDLE_REQUEST);

                        break;
                    case HANDLER_ERROR:
                        /* something went wrong */
                        connection_set_state(srv, con, CON_STATE_ERROR);
                        break;
                    default:
                        log_error_write(srv, __FILE__, __LINE__, "sdd", "unknown ret-value: ", con->fd, r);
                        break;
                }

                break;
            case CON_STATE_RESPONSE_START:
                /*
                 * the decision is done
                 * - create the HTTP-Response-Header
                 *
                 */

                if ( srv->srvconf.log_state_handling )
                {
                    log_error_write(srv, __FILE__, __LINE__, "sds",
                                    "state for fd", con->fd, connection_get_state(con->state));
                }

                if ( -1 == connection_handle_write_prepare(srv, con) )
                {
                    connection_set_state(srv, con, CON_STATE_ERROR);

                    break;
                }

                connection_set_state(srv, con, CON_STATE_WRITE);
                break;
            case CON_STATE_RESPONSE_END: /* transient */
                /* log the request */

                if ( srv->srvconf.log_state_handling )
                {
                    log_error_write(srv, __FILE__, __LINE__, "sds",
                                    "state for fd", con->fd, connection_get_state(con->state));
                }

                plugins_call_handle_request_done(srv, con);

                srv->con_written++;

                if ( con->keep_alive )
                {
                    connection_set_state(srv, con, CON_STATE_REQUEST_START);

#if 0
                    con->request_start = srv->cur_ts;
                    con->read_idle_ts = srv->cur_ts;
#endif
                }
                else
                {
                    switch ( r = plugins_call_handle_connection_close(srv, con) )
                    {
                        case HANDLER_GO_ON:
                        case HANDLER_FINISHED:
                            break;
                        default:
                            log_error_write(srv, __FILE__, __LINE__, "sd", "unhandling return value", r);
                            break;
                    }

#ifdef USE_OPENSSL
                    if ( srv_sock->is_ssl )
                    {
                        switch ( SSL_shutdown(con->ssl) )
                        {
                            case 1:
                                /* done */
                                break;
                            case 0:
                                /* wait for fd-event
                                 *
                                 * FIXME: wait for fdevent and call SSL_shutdown again
                                 *
                                 */

                                break;
                            default:
                                log_error_write(srv, __FILE__, __LINE__, "ss", "SSL:",
                                                ERR_error_string(ERR_get_error(), NULL));
                        }
                    }
#endif
                    if ( (0 == shutdown(con->fd, SHUT_WR)) )
                    {
                        con->close_timeout_ts = srv->cur_ts;
                        connection_set_state(srv, con, CON_STATE_CLOSE);
                    }
                    else
                    {
                        connection_close(srv, con);
                    }

                    srv->con_closed++;
                }

                connection_reset(srv, con);

                break;
            case CON_STATE_CONNECT:
                if ( srv->srvconf.log_state_handling )
                {
                    log_error_write(srv, __FILE__, __LINE__, "sds",
                                    "state for fd", con->fd, connection_get_state(con->state));
                }

                chunkqueue_reset(con->read_queue);

                con->request_count = 0;

                break;
            case CON_STATE_CLOSE:
                if ( srv->srvconf.log_state_handling )
                {
                    log_error_write(srv, __FILE__, __LINE__, "sds",
                                    "state for fd", con->fd, connection_get_state(con->state));
                }

                /* we have to do the linger_on_close stuff regardless
                 * of con->keep_alive; even non-keepalive sockets may
                 * still have unread data, and closing before reading
                 * it will make the client not see all our output.
                 */
                {
                    int len;
                    char buf[1024];

                    len = read(con->fd, buf, sizeof(buf));
                    if ( len == 0 || (len < 0 && errno != EAGAIN && errno != EINTR) )
                    {
                        con->close_timeout_ts = srv->cur_ts - (HTTP_LINGER_TIMEOUT+1);
                    }
                }

                if ( srv->cur_ts - con->close_timeout_ts > HTTP_LINGER_TIMEOUT )
                {
                    connection_close(srv, con);

                    if ( srv->srvconf.log_state_handling )
                    {
                        log_error_write(srv, __FILE__, __LINE__, "sd",
                                        "connection closed for fd", con->fd);
                    }
                }

                break;
            case CON_STATE_READ_POST:
            case CON_STATE_READ:
                if ( srv->srvconf.log_state_handling )
                {
                    log_error_write(srv, __FILE__, __LINE__, "sds",
                                    "state for fd", con->fd, connection_get_state(con->state));
                }

                connection_handle_read_state(srv, con);
                break;
            case CON_STATE_WRITE:
                if ( srv->srvconf.log_state_handling )
                {
                    log_error_write(srv, __FILE__, __LINE__, "sds",
                                    "state for fd", con->fd, connection_get_state(con->state));
                }

                /* only try to write if we have something in the queue */
                if ( !chunkqueue_is_empty(con->write_queue) )
                {
#if 0
                    log_error_write(srv, __FILE__, __LINE__, "dsd",
                                    con->fd,
                                    "packets to write:",
                                    con->write_queue->used);
#endif
                }
                if ( !chunkqueue_is_empty(con->write_queue) && con->is_writable )
                {
                    if ( -1 == connection_handle_write(srv, con) )
                    {
                        log_error_write(srv, __FILE__, __LINE__, "ds",
                                        con->fd,
                                        "handle write failed.");
                        connection_set_state(srv, con, CON_STATE_ERROR);
                    }
                    else if ( con->state == CON_STATE_WRITE )
                    {
                        con->write_request_ts = srv->cur_ts;
                    }
                }

                break;
            case CON_STATE_ERROR: /* transient */

                /* even if the connection was drop we still have to write it to the access log */
                if ( con->http_status )
                {
                    plugins_call_handle_request_done(srv, con);
                }
#ifdef USE_OPENSSL
                if ( srv_sock->is_ssl )
                {
                    int ret, ssl_r;
                    unsigned long err;
                    ERR_clear_error();
                    switch ( (ret = SSL_shutdown(con->ssl)) )
                    {
                        case 1:
                            /* ok */
                            break;
                        case 0:
                            ERR_clear_error();
                            if ( -1 != (ret = SSL_shutdown(con->ssl)) ) break;

                            /* fall through */
                        default:

                            switch ( (ssl_r = SSL_get_error(con->ssl, ret)) )
                            {
                                case SSL_ERROR_WANT_WRITE:
                                case SSL_ERROR_WANT_READ:
                                    break;
                                case SSL_ERROR_SYSCALL:
                                    /* perhaps we have error waiting in our error-queue */
                                    if ( 0 != (err = ERR_get_error()) )
                                    {
                                        do
                                        {
                                            log_error_write(srv, __FILE__, __LINE__, "sdds", "SSL:",
                                                            ssl_r, ret,
                                                            ERR_error_string(err, NULL));
                                        } while ( (err = ERR_get_error()) );
                                    }
                                    else if ( errno != 0 ) /* ssl bug (see lighttpd ticket #2213): sometimes errno == 0 */
                                    {
                                        log_error_write(srv, __FILE__, __LINE__, "sddds", "SSL (error):",
                                                        ssl_r, ret, errno,
                                                        strerror(errno));
                                    }

                                    break;
                                default:
                                    while ( (err = ERR_get_error()) )
                                    {
                                        log_error_write(srv, __FILE__, __LINE__, "sdds", "SSL:",
                                                        ssl_r, ret,
                                                        ERR_error_string(err, NULL));
                                    }

                                    break;
                            }
                    }
                }
                ERR_clear_error();
#endif

                switch ( con->mode )
                {
                    case DIRECT:
#if 0
                        log_error_write(srv, __FILE__, __LINE__, "sd",
                                        "emergency exit: direct",
                                        con->fd);
#endif
                        break;
                    default:
                        switch ( r = plugins_call_handle_connection_close(srv, con) )
                        {
                            case HANDLER_GO_ON:
                            case HANDLER_FINISHED:
                                break;
                            default:
                                log_error_write(srv, __FILE__, __LINE__, "sd", "unhandling return value", r);
                                break;
                        }
                        break;
                }

                connection_reset(srv, con);

                /* close the connection */
                if ( (0 == shutdown(con->fd, SHUT_WR)) )
                {
                    con->close_timeout_ts = srv->cur_ts;
                    connection_set_state(srv, con, CON_STATE_CLOSE);

                    if ( srv->srvconf.log_state_handling )
                    {
                        log_error_write(srv, __FILE__, __LINE__, "sd",
                                        "shutdown for fd", con->fd);
                    }
                }
                else
                {
                    connection_close(srv, con);
                }

                con->keep_alive = 0;

                srv->con_closed++;

                break;
            default:
                log_error_write(srv, __FILE__, __LINE__, "sdd",
                                "unknown state:", con->fd, con->state);

                break;
        }

        if ( done == -1 )
        {
            done = 0;
        }
        else if ( ostate == con->state )
        {
            done = 1;
        }
    }

    if ( srv->srvconf.log_state_handling )
    {
        log_error_write(srv, __FILE__, __LINE__, "sds",
                        "state at exit:",
                        con->fd,
                        connection_get_state(con->state));
    }

    switch ( con->state )
    {
        case CON_STATE_READ_POST:
        case CON_STATE_READ:
        case CON_STATE_CLOSE:
            fdevent_event_set(srv->ev, &(con->fde_ndx), con->fd, FDEVENT_IN);
            break;
        case CON_STATE_WRITE:
            /* request write-fdevent only if we really need it
             * - if we have data to write
             * - if the socket is not writable yet
             */
            if ( !chunkqueue_is_empty(con->write_queue) &&
                 (con->is_writable == 0) &&
                 (con->traffic_limit_reached == 0) )
            {
                fdevent_event_set(srv->ev, &(con->fde_ndx), con->fd, FDEVENT_OUT);
            }
            else
            {
                fdevent_event_del(srv->ev, &(con->fde_ndx), con->fd);
            }
            break;
        default:
            fdevent_event_del(srv->ev, &(con->fde_ndx), con->fd);
            break;
    }

    return 0;
}
